﻿#region License
/*
Copyright 2014 Christian Thurmann-Nielsen

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/
#endregion

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Diagnostics;
using System.Globalization;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using EnvDTE;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.Shell;
using System.ComponentModel.Design;
using QuickGraph;
using GraphSharp;
using Parser;
using NTNU.DependencySimulator.GUI;

namespace NTNU.DependencySimulator
{
    public delegate IBidirectionalGraph<object, IEdge<object>> SolutionAnalyzedCallback();
    /// <summary>
    /// This class implements the tool window exposed by this package and hosts a user control.
    ///
    /// In Visual Studio tool windows are composed of a frame (implemented by the shell) and a pane, 
    /// usually implemented by the package implementer.
    ///
    /// This class derives from the ToolWindowPane class provided from the MPF in order to use its 
    /// implementation of the IVsUIElementPane interface.
    /// </summary>
    [Guid("0049aa61-781d-4dbe-8e39-3daf0746efcc")]
    public class DependencySimulatorWindow : ToolWindowPane
    {
        private EnvDTE.DTE dte;
        public IBidirectionalGraph<object, IEdge<object>> solGraph;
        //public extern void Parser.ProjectParser.parse(SolutionAnalyzedCallback callback, string solPath);
        public string solPath;
        private MyControl control = null;
        private string[] dropDownComboChoices = { DependencySimulatorResurces.Move, DependencySimulatorResurces.AddEdge, DependencySimulatorResurces.RemoveEdge };
        private string currentDropDownComboChoice = DependencySimulatorResurces.Move;
        ClassFileFinder cf;
        IVsObjectManager2 ivsom2;

        /// <summary>
        /// Standard constructor for the tool window.
        /// </summary>
        public DependencySimulatorWindow() :
            base(null)
        {
            // This is the user control hosted by the tool window 
            control = new MyControl();
            base.Content = control; 
            dte = (EnvDTE.DTE) GetService(typeof(EnvDTE.DTE));
            // Set the window title reading it from the resources. 
            this.Caption = DependencySimulatorResurces.ToolWindowTitle;
            // Set the image that will appear on the tab of the window frame 
            // when docked with an other window 
            // The resource ID correspond to the one defined in the resx file 
            // while the Index is the offset in the bitmap strip. Each image in 
            // the strip being 16x16. 
            this.BitmapResourceID = 301;
            this.BitmapIndex = 1;

            // Create the toolbar. 
            this.ToolBar = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                PkgCmdIDList.ToolbarID);
            this.ToolBarLocation = (int)VSTWT_LOCATION.VSTWT_TOP;


            OleMenuCommandService mcs =
                    GetService(typeof(IMenuCommandService)) as OleMenuCommandService;

            // Create the handlers for the toolbar commands. 
            if (null != mcs)
            {
                // Add the load file button
                var btnLoadFile = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidLoadFile);
                var menuItemLoadFile = new MenuCommand(new EventHandler(LoadFile),
                                                btnLoadFile);
                mcs.AddCommand(menuItemLoadFile);

                // Add the undo button
                var btnUndo = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidUndo);
                var menuItemUndo = new MenuCommand(new EventHandler(Undo),
                                                btnUndo);
                mcs.AddCommand(menuItemUndo);

                // Add the redo button
                var btnRedo = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidRedo);
                var menuItemRedo = new MenuCommand(new EventHandler(Redo),
                                                btnRedo);
                mcs.AddCommand(menuItemRedo);

                // Add the refresh button
                var btnRecalculate = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidRecalculate);
                var menuItemRecalculate = new MenuCommand(new EventHandler(Recalculate),
                                                btnRecalculate);
                mcs.AddCommand(menuItemRecalculate);


                // Add the simulate refactoring button
                var btnSimulate = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidSimulateRefactoring);
                var menuItemSimulate = new MenuCommand(new EventHandler(SimulateRefactoring),
                                                btnSimulate);
                mcs.AddCommand(menuItemSimulate);

                // Add the report button
                var btnMapFiles = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidMapFiles);
                var menuItemMapFiles = new MenuCommand(new EventHandler(MapFiles),
                                                btnMapFiles);
                mcs.AddCommand(menuItemMapFiles);

                // Add the show result button
                var btnResult = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidShowResult);
                var menuItemResult = new MenuCommand(new EventHandler(ShowResult),
                                                btnResult);
                mcs.AddCommand(menuItemResult);

                // Add the show report button
                var btnReport = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidReport);
                var menuItemReport = new MenuCommand(new EventHandler(ShowReport),
                                                btnReport);
                mcs.AddCommand(menuItemReport);

                // Add the change view button
                var btnChangeView = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidChangeView);
                var menuItemChangeView = new MenuCommand(new EventHandler(ChangeView),
                                                btnChangeView);
                mcs.AddCommand(menuItemChangeView);

                // Add the help button
                var btnHelp = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidHelp);
                var menuItemHelp = new MenuCommand(new EventHandler(ShowHelp),
                                                btnHelp);
                mcs.AddCommand(menuItemHelp);

                // Add the rollback button
                var btnRollback = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidRollback);
                var menuItemRollback = new MenuCommand(new EventHandler(Rollback),
                                                btnRollback);
                mcs.AddCommand(menuItemRollback);

                // Add the rollback button
                var btnFitness = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                        PkgCmdIDList.cmdidFitness);
                var menuItemFitness = new MenuCommand(new EventHandler(Fitness),
                                                btnFitness);
                mcs.AddCommand(menuItemFitness);

                // --- Initialize the DropDownCombo
                var menuMyDropDownComboCommandID = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                                    (int)PkgCmdIDList.cmdidDropDown);
                var menuMyDropDownComboCommand = new OleMenuCommand(new EventHandler(OnMenuMyDropDownCombo),
                                                                        menuMyDropDownComboCommandID);
                menuMyDropDownComboCommand.ParametersDescription = "$";
                mcs.AddCommand(menuMyDropDownComboCommand);

                // --- Initialize the “GetList” command for DropDownCombo
                var menuMyDropDownComboGetListCommandID = new CommandID(GuidList.guidDepToolWIndowCmdSet,
                                                            (int)PkgCmdIDList.cmdidMyDropDownComboGetList);
                var menuMyDropDownComboGetListCommand = new OleMenuCommand(new EventHandler(OnMenuMyDropDownComboGetList),
                  menuMyDropDownComboGetListCommandID);
                mcs.AddCommand(menuMyDropDownComboGetListCommand);
            }
        }
        public DependencySimulatorWindow(string dir)
            : this()
        {
            LoadFile();
        }
        private void ShowReport(object sender, EventArgs e)
        {
            control.ShowReport();
        }
        private void ShowResult(object sender, EventArgs e)
        {
            control.ShowResult();
        }
        private void ShowHelp(object sender, EventArgs e)
        {
            control.ShowHelp();
        }
        private void ChangeView(object sender, EventArgs e)
        {
            control.ChangeView();
        }
        private void Fitness(object sender, EventArgs e)
        {
            control.ShowFitness();
        }
        private void SimulateRefactoring(object sender, EventArgs e)
        {
            control.SimulateRefactoring();
        }

        private void Recalculate(object sender, EventArgs e)
        {
            control.Recalculate();
        }

        private void Rollback(object sender, EventArgs e)
        {
            control.Rollback();
        }

        private void MapFiles(object sender, EventArgs e)
        {
            control.WriteMapsToFile();
        }

        private void Redo(object sender, EventArgs e)
        {
            control.Redo();
        }

        private void Undo(object sender, EventArgs e)
        {
            control.Undo();
        }

        public string CurrentDropDownComboChoice
        {
            get { return currentDropDownComboChoice; }
            set
            {
                currentDropDownComboChoice = value;
                if (value == DependencySimulatorResurces.AddEdge) { control.SetSelectedAction(1); }
                else if (value == DependencySimulatorResurces.RemoveEdge) { control.SetSelectedAction(2); }
                else { control.SetSelectedAction(3); }
            }
        } 
        private void OnMenuMyDropDownComboGetList(object sender, EventArgs e)
        {
            if ((null == e) || (e == EventArgs.Empty))
            {
                // --- We should never get here; EventArgs are required.
                throw (new ArgumentNullException(DependencySimulatorResurces.EventArgsRequired));
            }
            OleMenuCmdEventArgs eventArgs = e as OleMenuCmdEventArgs;
            if (eventArgs != null)
            {
                object inParam = eventArgs.InValue;
                IntPtr vOut = eventArgs.OutValue;
                if (inParam != null)
                {
                    throw (new ArgumentException(DependencySimulatorResurces.InParamIllegal));
                }
                else if (vOut != IntPtr.Zero)
                {
                    Marshal.GetNativeVariantForObject(this.dropDownComboChoices, vOut);
                }
                else
                {
                    throw (new ArgumentException(DependencySimulatorResurces.OutParamRequired));
                }
            }
        }

        private void OnMenuMyDropDownCombo(object sender, EventArgs e)
        {
            if (e == EventArgs.Empty)
            {
                throw (new ArgumentException(DependencySimulatorResurces.EventArgsRequired));
            }
            OleMenuCmdEventArgs eventArgs = e as OleMenuCmdEventArgs;
            if (eventArgs != null)
            {
                string newChoice = eventArgs.InValue as string;
                IntPtr vOut = eventArgs.OutValue;
                if (vOut != IntPtr.Zero && newChoice != null)
                {
                    throw (new ArgumentException(DependencySimulatorResurces.BothParamIllegal));
                }
                else if (vOut != IntPtr.Zero)
                {
                    Marshal.GetNativeVariantForObject(this.currentDropDownComboChoice, vOut);
                }
                else if (newChoice != null)
                {
                    bool validInput = false;
                    int indexInput = -1;
                    for (indexInput = 0; indexInput < dropDownComboChoices.Length; indexInput++)
                    {
                        if (String.Compare(dropDownComboChoices[indexInput], newChoice,
                          StringComparison.CurrentCultureIgnoreCase) == 0)
                        {
                            validInput = true;
                            break;
                        }
                    }
                    if (validInput)
                    {
                        CurrentDropDownComboChoice = dropDownComboChoices[indexInput];
                        //System.Windows.MessageBox.Show("Current choice: " + this.CurrentDropDownComboChoice);
                    }
                    else
                    {
                        throw (new ArgumentException(DependencySimulatorResurces.InParamIllegal));
                    }
                }
                else
                {
                    throw (new ArgumentException("Param cant be NULL"));
                }
            }
            else
            {
                throw (new ArgumentException(DependencySimulatorResurces.EventArgsRequired));
            }
        }
        public void StartFromDir()
        {
            LoadFile();
        }
        private void LoadFile()
        {
            dte = (EnvDTE.DTE)GetService(typeof(EnvDTE.DTE));
            
            //IVsLibrary2 l;
            //ivsom2.FindLibrary(s,out l);
            try
            {
                if (dte != null)
                {
                    ivsom2 = (IVsObjectManager2)GetService(typeof(SVsObjectManager));
                    cf = new ClassFileFinder(dte.Solution.Projects);
                    control.DTE = dte;
                    string soldir = System.IO.Path.GetDirectoryName(dte.Solution.FullName);
                    control.GenerateGraph(cf,soldir);
                }
                else
                    throw new ArgumentException();
            }
            catch (ArgumentException)
            {
                System.Windows.MessageBox.Show("No solution is loaded into Visual Studio");
            }
        }
        private void LoadFile(object sender, EventArgs arguments)
        {
            LoadFile();
        }
        private void DropDownHandler(object sender, EventArgs arguments)
        {
        }
        
        private void OnBeforeQueryStatus(object sender, EventArgs arguments)
        {
            //var selectedItems = dte.SelectedItems.Item;
            //bool fileTypeSupported = selectedItems.Any((si) => si.Name.EndsWith(".js"));
            //Parser.ProjectParser p;
            //(UIHierarchy)ideObject.Windows.Item(EnvDTE.Constants.vsWindowKindSolutionExplorer).Object obj;
            //SolutionAnalyzedCallback callback = new SolutionAnalyzedCallback(MyToolWindow.onAnalyzedSolutionComplete);
            //Parser.ProjectParser.parse(callback, solPath);
        }
        public static IBidirectionalGraph<object, IEdge<object>> onAnalyzedSolutionComplete()
        {
            IBidirectionalGraph<object, IEdge<object>> graph = null;
            return graph;
        }
    }
}
