﻿#region License
/*
Copyright 2014 Christian Thurmann-Nielsen

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/
#endregion

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Threading;
using QuickGraph;
using GraphController;
using GraphController.Common;
using GraphX.Controls;
using Parser;
using NTNU.DependencySimulator.GUI;
using GraphController.GUI;
using GraphController.Utility;
using Microsoft.VisualStudio.Shell.Interop;

namespace NTNU.DependencySimulator
{
    /// <summary>
    /// Interaction logic for MyControl.xaml
    /// </summary>
    public partial class MyControl : UserControl, INotifyPropertyChanged
    {
        private static readonly DependencyProperty ProgressProperty =
            DependencyProperty.Register("Progress", typeof(double), typeof(MyControl));
        private Guid csLibGuid = new Guid("58f1bad0-2288-45b9-ac3a-d56398f7781d");
        private IVsObjectManager2 ivs;
        private IVsLibrary2 cslib;
        private IVsObjectList2 objlist;
        private string solutionDir;
        private BackgroundWorker worker;
        private MainController model;
        private List<DataGridColumn> _hiddenColumns;
        private ContextMenu cxMenu;
        private Dispatcher UIDispatcher;
        private DependencyFinder _finder;
        private EnvDTE.DTE _dte;
        private DependencyVertex _selectedVertex;
        
        public MyControl()
        {
            
            UIDispatcher = this.Dispatcher;
            model = new MainController();
            InitBackgroundWorker();
            graphLayout = new MyGraphLayout();
            graphLayout.Graph = GraphToVisualize;
            graphLayout.ShouldLayout = true;
            dataView = new DataGrid();
            _hiddenColumns = new List<DataGridColumn>();
            zoomControl = new ZoomControl();
            //zoomControl.AnimationLength = TimeSpan.Zero;
            InitializeComponent();
            
            this.Focus();
            Tracer.WriteTrace("MyControl(): focus: {0}", this.IsFocused);
        }

        public EnvDTE.DTE DTE
        {
            set { _dte = value; }
        }
        public DependencyGraph GraphToVisualize
        {
            get { return model.GraphToVisualize; }
            set
            {
                if (!Equals(value, model.GraphToVisualize))
                {
                    model.GraphToVisualize = value;
                    //model.GraphToVisualize.Analyse();
                    NotifyPropertyChanged("GraphToVisualize");
                }
            }
        }
        public ObservableCollection<DependencyVertex> ClassCollection
        {
            get { return model.ClassCollection; }
            set
            {
                model.ClassCollection = value;
                NotifyPropertyChanged("ClassCollection");
            }
        }
        private void InitBackgroundWorker()
        {
            worker = new BackgroundWorker();

            worker.WorkerSupportsCancellation = true;
            worker.WorkerReportsProgress = true;

            worker.DoWork += new DoWorkEventHandler(bw_DoWork);
            worker.ProgressChanged += new ProgressChangedEventHandler(bw_ProgressChanged);
            worker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(bw_RunWorkerCompleted);
        }
        private void UpdateView()
        {
            Tracer.WriteTrace("Load Complete: {0}", model.GraphToVisualize.Vertices.Count());
            model.FinishedLoading(); 

            graphLayout.ShouldLayout = true;
            graphLayout.Relayout();
            graphLayout.Graph = model.GraphToVisualize; 
            graphLayout.ShouldLayout = false;

            dataView.UpdateLayout();
            
            model.ReLayout();
        }
        private void bw_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            //pf.Hide();
            //UpdateView();
        }

        private void bw_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            // Checking if this thread has access to the object.
            if (UIDispatcher.CheckAccess())
            {
                // This thread has access so it can update the UI thread.      
                UpdateView();
            }
            else
            {
                // This thread does not have access to the UI thread.
                // Place the update method on the Dispatcher of the UI thread.
                UIDispatcher.BeginInvoke(DispatcherPriority.Normal,
                    (Action)(() => { UpdateView(); }));
            }
        }

        private void bw_DoWork(object sender, DoWorkEventArgs e)
        {
            //GenerateGraph();
            worker.ReportProgress(100,null);
        }

        public event PropertyChangedEventHandler PropertyChanged;

        public void RaisePropChanged(string name)
        {
            var eh = this.PropertyChanged;
            if (eh != null)
            {
                eh(this, new PropertyChangedEventArgs(name));
            }
        }
        public void SetSelectedAction(int i)
        {
            model.SetSelectedAction(i);
            if (model.SelectedVertex != null)
            {
                model.SelectedVertex.IsHighlighted = false;
                model.SelectedVertex = null;
            } 
        }
        private bool CheckInitialization()
        {
            if (model.IsInitialized)
                return true;
            else
            {
                MessageBox.Show("No Solution loaded into Visual Studio");
                return false;
            }
        }
        private bool CheckClassView()
        {
            if (!CheckInitialization())
                return false;
            if (model.IsClassView)
                return true;
            else
            {
                MessageBox.Show("Not Supported");
                return false;
            }
        }
        public void Undo()
        {
            if (CheckClassView())
            {
                model.UndoAction();
                dataView.UpdateLayout();
            }
        }
        public void Redo()
        {
            if (CheckClassView())
            {
                model.RedoAction();
                dataView.UpdateLayout();
            }
            
        }

        public void GenerateGraph(ClassFileFinder cf,string dir)
        {
            solutionDir = dir;
            Tracer.WriteTrace("OnButton, file: {0}, {1}", solutionDir, model.IsInitialized);
            ivs = DependencySimulatorPackage.GetGlobalService(typeof(SVsObjectManager)) as IVsObjectManager2;

            model.ReInit();

            model.InitEventHandler();
            _finder = new DependencyFinder( model, UIDispatcher);
            ClassCollection = model.ClassCollection;
            //worker.RunWorkerAsync();
            List<ClassFile> list = cf.GetClassFiles();

            try
            {
                _finder.parseClassFiles(list);
            }
            catch(System.IO.FileLoadException)
            {
                System.Windows.MessageBox.Show("No solution is loaded into Visual Studio");
                model.DisposeProgressBar();
                return;
            }
            catch(System.Exception e)
            {
                MessageBox.Show("Something went wrong. Please try again.\n\n" + e.StackTrace);
                model.DisposeProgressBar();
                return;
            }

            UpdateView();
            Tracer.WriteTrace("OnButton, init: {0}", model.IsInitialized);
        }

        public void LoadSolution(ClassFileFinder cf)
        {
            //System.Windows.Forms.OpenFileDialog ofd = new System.Windows.Forms.OpenFileDialog();
            //ofd.Filter = "Solution Files (.sln)|*.sln|All Files ()|*.*";

            //if (ofd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            //{
            //    solutionDir = System.IO.Path.GetDirectoryName(ofd.FileName);
            //    GenerateGraph();
            //    //worker.RunWorkerAsync();
            //    UpdateView();
            //    //model.PrintMaps();
            //    //graphLayout.Relayout();
            //    //NotifyPropertyChanged("Graph");
            //}
        }
        
        internal void SimulateRefactoring()
        {
            //MessageBox.Show(model.GenerateReport());
            //model.WriteToFile();
            if (model.IsInitialized)
            {
                ParameterInputView v = new ParameterInputView(model);
                v.Show();
            }
            else
                MessageBox.Show("No Solution has been loaded");
           
        }
        internal void WriteMapsToFile()
        {
            if(solutionDir != null)
                model.WriteReport(solutionDir);
        }
        internal void Recalculate()
        {
            if (!model.IsInitialized)
            {
                MessageBox.Show("No solution loaded into Visual Studio");
                return;
            } 
            if (!model.IsClassView)
            {
                MessageBox.Show("Not supported in fileview");
                return;
            }
            DependencyGraph g = model.Recalculate();
            if (g == null)
                return;
            
            graphLayout.ShouldLayout = true;
            graphLayout.Graph = g;
            graphLayout.ShouldLayout = false; 
            ClassCollection = model.RecalculateClassCollection();
            //dataView.UpdateLayout();
        }
        internal void ShowReport()
        {
            if(CheckInitialization())
            {
                var edges = model.GetReport();
                if (edges != null)
                {
                    ReportViewer c = new ReportViewer(edges, solutionDir);
                    c.Show();
                }
                else
                    MessageBox.Show("No report available!");
            }
            
        }
        internal void Rollback()
        {
            CheckRollback(model.Rollback);
        }
        
        internal void RollbackPrevious()
        {
            CheckRollback(model.RollbackPreviousResult);
        }
        private void CheckRollback(Func<DependencyGraph> a)
        {
            if (!model.IsInitialized)
            {
                MessageBox.Show("No solution loaded into Visual Studio");
                return;
            }
            if (!model.IsClassView)
            {
                a();
                return;
            }
            DependencyGraph g = a();
            if (g == null)
                return;

            graphLayout.ShouldLayout = true;
            graphLayout.Graph = g;
            graphLayout.ShouldLayout = false;
            ClassCollection = model.ClassCollection;
        }
        internal void AddResultToModel(SimulationResult r)
        {
            Recalculate();
            model.AddResult(r);
        }
        internal void ShowResult()
        {
            if(CheckInitialization())
            {
                if (model.GetCurrentResult() != null)
                {
                    SimulationResultViewer s = new SimulationResultViewer(model.GetCurrentResult(), model);
                    s.Show();
                }
                else
                    MessageBox.Show("No result available!");
            }
        }
        internal void ShowHelp()
        {
            HelpView v = new HelpView();
            v.Show();
        }
        internal void ChangeView()
        {
            if(model.IsInitialized)
            {
                graphLayout.ShouldLayout = true;
                graphLayout.Graph = model.ChangeView();
                //Throws KeyNotFoundException on next edge manipulation.
                ClassCollection = model.ClassCollection;
                //dataView.UpdateLayout();
                graphLayout.ShouldLayout = false;
            }
        }
        private void ShowInputDialog(string comment)
        {
            
            InputBoxResult test = InputBox.Show(comment + "Write the name of the new node" + "\n"
                  , "Add new node", "", 100, 0);

            if (test.ReturnCode == System.Windows.Forms.DialogResult.OK)
            {
                Tracer.WriteTrace("MenuItem_AddNode onClick() -> OK");
                model.AddNodeFromGUI(test.Name,test.Namespace);
                dataView.UpdateLayout();
            }
            else if (test.ReturnCode == System.Windows.Forms.DialogResult.Retry)
            {
                ShowInputDialog("Name cannot be empty, please retry. \n");
            }
        }
        /**
         *      METHODS FOR CLICKING ON MENU ITEMS
         * 
         */

        
        private void MenuItem_AddNode(object sender, RoutedEventArgs e)
        {
            if (!CheckClassView())
            {
                return;
            }
            Tracer.WriteTrace("MenuItem_AddNode onClick()");
            var menuItem = sender as MenuItem;
            if (menuItem != null && menuItem.Tag != null)
            {
                ShowInputDialog("");
                
            }
        }

        private void MenuItem_AddEdge(object sender, RoutedEventArgs e)
        {

        }


        private void MenuItem_EdgeInfo(object sender, RoutedEventArgs e)
        {
            try
            {
                MenuItem m = (MenuItem)e.OriginalSource;
                Tracer.WriteTrace("Remove edge: {0}, {1}", sender.GetType().ToString(), m.DataContext.ToString());
                DependencyEdge data = (DependencyEdge)m.DataContext;
                
                CodeViewer c = new CodeViewer("Edge Info",data.GetEdgeInfo());
                c.Show();
            }
            catch (InvalidCastException)
            {
                return;
            }
        }
        private void MenuItem_RemoveEdge(object sender, RoutedEventArgs e)
        {
            if (!CheckClassView())
            {
                return;
            }
            try
            {
                MenuItem m = (MenuItem)e.OriginalSource;
                DependencyEdge data = (DependencyEdge)m.DataContext;
                Tracer.WriteTrace("Remove edge: {0}, {1}", sender.GetType().ToString(), m.DataContext.ToString());
                if (data.IsCyclic)
                {
                    string question = "This Edge is cyclic! Do you want to remove the edge from: " + data.Source.FullName + "(source) to " + data.Target.FullName + "(target)?"
                                        + "\n(Answer No for vice versa or cancel for none of these edges.)";
                    MessageBoxResult result = MessageBox.Show(question, "Cyclic edge",
                                                                                System.Windows.MessageBoxButton.YesNoCancel, System.Windows.MessageBoxImage.Question);

                    if (result == MessageBoxResult.Yes)
                    {
                        model.PerformRemoveEdgeAction(data);
                    }
                    else if (result == MessageBoxResult.No)
                    {
                        model.PerformRemoveEdgeAction(data.CyclicSibling);
                    }
                }
                else
                    model.PerformRemoveEdgeAction(data);
            }
            catch (InvalidCastException)
            {
                return;
            }
            catch(KeyNotFoundException)
            {
                return;
            }
        }
        private void MenuItem_SetMainClass(object sender, RoutedEventArgs e)
        {
            if (!CheckClassView())
            {
                return;
            }
            MenuItem m;
            DependencyVertex data;
            try
            {

                Tracer.WriteTrace("remove node: {0}", e.OriginalSource.ToString());
                m = (MenuItem)e.OriginalSource;
                data = (DependencyVertex)m.DataContext;
                if (data != null)
                {
                    model.SetMainClass(data.FullName);
                }
            }
            catch (InvalidCastException)
            {
                if (_selectedVertex != null)
                    model.SetMainClass(_selectedVertex.FullName);
                return;
            }

        }
        private void MenuItem_RemoveNode(object sender, RoutedEventArgs e)
        {
            if (!CheckClassView())
            {
                return;
            }
            MenuItem m;
            DependencyVertex data;
            try
            {
                
                Tracer.WriteTrace("remove node: {0}",e.OriginalSource.ToString());
                m = (MenuItem)e.OriginalSource;
                data = (DependencyVertex)m.DataContext;
                if (data != null)
                {
                    model.PerformRemoveNodeAction(data);
                    dataView.UpdateLayout();
                }
            }
            catch (InvalidCastException)
            {
                if (_selectedVertex != null)
                {
                    model.PerformRemoveNodeAction(_selectedVertex);
                    dataView.UpdateLayout();
                }
                return;
            }
            
        }

        private void MenuItem_EditReferences(object sender, RoutedEventArgs e)
        {
            if (!CheckClassView())
            {
                return;
            }
            MenuItem m;
            DependencyEdge data;
            try
            {
                Tracer.WriteTrace("remove node: {0}", e.OriginalSource.ToString()); 
                m = (MenuItem)e.OriginalSource;
                data = (DependencyEdge)m.DataContext;
                if (data != null)
                {
                    if(data.IsCyclic)
                    {   
                        string question = "This Edge is cyclic! Do you want to edit references from: " +  data.Source.FullName + "(source) to " + data.Target.FullName + "(target)?"
                                            + "\n(Answer No for vice versa or cancel for none of these edges.)";
                        MessageBoxResult result = MessageBox.Show(question, "Cyclic edge", 
                                                                                    System.Windows.MessageBoxButton.YesNoCancel, System.Windows.MessageBoxImage.Question);
                        
                        if(result == MessageBoxResult.Yes)
                        {
                            RemoveReferenceViewer v = new RemoveReferenceViewer(data, model);
                            v.Show();
                        }
                        else if(result == MessageBoxResult.No)
                        {
                            RemoveReferenceViewer v = new RemoveReferenceViewer(data.CyclicSibling, model);
                            v.Show();
                        }
                        
                        
                    }
                    else
                    {
                        RemoveReferenceViewer v = new RemoveReferenceViewer(data, model);
                        v.Show();
                    }
                    
                }
            }
            catch (InvalidCastException)
            {
                return;
            }
            
        }
        private void Border_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            if (!model.IsClassView)
            {
                return;
            }
            if (model.IsActionMove())
                return;

            TextBlock b;
            DependencyVertex data;
            try
            {
                b = (TextBlock)e.OriginalSource;
                data = (DependencyVertex)b.DataContext;
            }
            catch (InvalidCastException)
            {
                return;
            }
             
            Tracer.WriteTrace("Border onClick() ->{0}, {1}", sender.ToString(), data.FullName);
            model.DoAction(data);
            dataView.UpdateLayout();
        }

        private void Border_MouseRightButtonDown(object sender, MouseButtonEventArgs e)
        {
            Tracer.WriteTrace("Edge, onRightClick()");
        }

        private void CommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
        {
        }

        private void ZoomControl_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            Tracer.WriteTrace("Zoom onCLick()");
            //model.SelectedVertex = null;
            /*if (model.SelectedVertex != null)
            {
                model.SelectedVertex.IsHighlighted = false;
                model.SelectedVertex = null;
            } */
        }

        private void ZoomControl_MouseDoubleClick(object sender, MouseButtonEventArgs e)
        {
            if (model.SelectedVertex != null)
            {
                model.SelectedVertex.IsHighlighted = false;
                model.SelectedVertex = null;
            }
        }

        private void listView_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            DependencyObject depObj = (DependencyObject)e.OriginalSource;
            Tracer.WriteTrace("Left click: {0} , {1}", e.OriginalSource.ToString(), depObj.ToString());

            if (depObj is ScrollViewer)
            {
                if (model.SelectedVertex != null)
                {
                    model.SelectedVertex.IsHighlighted = false;
                }
               
                return;
            }
            DependencyVertex v;
            try
            {
                v = (DependencyVertex)dataView.SelectedItem;
                Tracer.WriteTrace("selected item: {0}", v.Name);

                //graphLayout.UnHighlightVertex(model.SelectedVertex);
                if (model.SelectedVertex != null) model.SelectedVertex.IsHighlighted = false;
                v.IsHighlighted = true;
                graphLayout.UpdateLayout();

                model.SelectedVertex = v;
            }
            catch (Exception)
            {
            }
            //MyVertexBase v = (MyVertexBase)listView.SelectedItem;
        }

        private void listView_MouseDoubleClick(object sender, MouseButtonEventArgs e)
        {
            DependencyObject depObj = (DependencyObject)e.OriginalSource;
            Tracer.WriteTrace("Left click: {0} , {1}", e.OriginalSource.ToString(), depObj.ToString());

            if (depObj is ScrollViewer)
            {
                model.SelectedVertex.IsHighlighted = false;
            }
            DependencyVertex v;
            try
            {
                v = (DependencyVertex)dataView.SelectedItem;
                Tracer.WriteTrace("selected item: {0}", v.Name);
                
                //graphLayout.UnHighlightVertex(model.SelectedVertex);
                if( model.SelectedVertex != null )model.SelectedVertex.IsHighlighted = false;
                v.IsHighlighted = true;
                graphLayout.UpdateLayout();
                
                model.SelectedVertex = v;
            }
            catch (Exception)
            {
            }
        }


        private void DataGrid_GenerateColumns(object sender, EventArgs e)
        {
            Tracer.WriteTrace("Create context menu for DataGrid: {0}", dataView.Columns.Count);
            cxMenu = new ContextMenu();
            List<DataGridColumn> columnsToRemove = new List<DataGridColumn>();

            string[] accepted = { "Name", "Namespace", "FanOut", "FanIn", "SccCount", "Reachability", "MethodCount", "StaticMethodCount", 
                                    "Constructors", "IsInterface", "IsEnum", "Fields", "Properties","AreAllMethodsStatic", "AreAllFieldsStatic", 
                                    "StaticFields", "PublicStaticFields", "PrivateFields", "Project", "ProjectDependency", "DependsOnSCC" };
            string[] hidden = { "Namespace", "SccCount", "AreAllMethodsStatic", "AreAllFieldsStatic", "StaticFields", "StaticMethodCount", "DependsOnSCC",
                                  "Constructors", "PublicStaticFields", "Properties", "PrivateFields", "IsInterface", "IsEnum", "Project", "ProjectDependency" };
            foreach (DataGridColumn c in dataView.Columns)
            {
                if (!accepted.Contains(c.Header.ToString()))
                {
                    Tracer.WriteTrace("removing column: {0}",c.Header.ToString());
                    columnsToRemove.Add(c);
                    continue;
                }
                c.IsReadOnly = true;
                MenuItem m = new MenuItem();
                m.Header = c.Header;
                m.IsChecked = true;
                cxMenu.Items.Add(m);
                m.Click += new RoutedEventHandler(DataGridColumn_Click);
                m.Checked += new RoutedEventHandler(DataGridColumn_Checked);
                m.Unchecked += new RoutedEventHandler(DataGridColumn_Unchecked);
                
                if (hidden.Contains(c.Header.ToString()) && !_hiddenColumns.Contains(c))
                {
                    int index = dataView.Columns.IndexOf(c);
                    m.IsChecked = false;
                    dataView.Columns[index].Visibility = System.Windows.Visibility.Hidden;
                    _hiddenColumns.Add(c);
                }
            }
            foreach (DataGridColumn c in columnsToRemove)
            {
                dataView.Columns.Remove(c);
            }
        }

        private void DataGridColumn_Unchecked(object sender, RoutedEventArgs e)
        {
            MenuItem m = sender as MenuItem;
            foreach (DataGridColumn c in dataView.Columns)
            {
                if( c.Header.ToString() == m.Header.ToString() )
                {
                    int index = dataView.Columns.IndexOf(c);
                    dataView.Columns[index].Visibility = System.Windows.Visibility.Hidden;
                    _hiddenColumns.Add(c);
                    break;
                }
            }
        }

        private void DataGridColumn_Checked(object sender, RoutedEventArgs e)
        {
            MenuItem m = sender as MenuItem;
            foreach (DataGridColumn c in _hiddenColumns)
            {
                if (c.Header.ToString() == m.Header.ToString())
                {
                    try
                    {
                        int index = dataView.Columns.IndexOf(c);
                        dataView.Columns[index].Visibility = System.Windows.Visibility.Visible;
                        _hiddenColumns.Remove(c);
                        Tracer.WriteTrace("Adding column: {0}", c.Header.ToString());
                        break;
                    }
                    catch(ArgumentOutOfRangeException)
                    {
                        continue;
                    }
                    
                }
            }
        }

        private void DataGridColumn_Click(object sender, RoutedEventArgs e)
        {
            MenuItem m = sender as MenuItem;
            if (m.IsChecked)
                m.IsChecked = false;
            else
                m.IsChecked = true;
        }
        private void DataGrid_RightClick(object sender, MouseButtonEventArgs e)
        {
            DependencyObject depObj = (DependencyObject)e.OriginalSource;

            while(depObj != null)
            {
                if (!CheckRightClickParent(depObj))
                    depObj = VisualTreeHelper.GetParent(depObj);
                else
                    break;
            }
            //Tracer.WriteTrace("Right click: {0} , {1}",e.OriginalSource.ToString(),depObj.ToString());
            //if(depObj is TextBlock)
            //{
                
            //}

            //if (depObj is ScrollViewer)
            //{
            //    ScrollViewer p = depObj as ScrollViewer;
            //    p.ContextMenu = cxMenu;
            //}

            //while (depObj != null && !(depObj is DataGridColumnHeader) )
            //    depObj = VisualTreeHelper.GetParent(depObj);

            //if (depObj == null)
            //    return;
            //Tracer.WriteTrace("     {0}", depObj.ToString());
            
            //if (depObj is DataGridColumnHeader)
            //{
            //    DataGridColumnHeader dgColHeader = depObj as DataGridColumnHeader;
            //    dgColHeader.ContextMenu = cxMenu;
            //    return;
            //}
        }
        private bool CheckRightClickParent(DependencyObject obj)
        {
            if(obj is TextBlock)
            {
                try
                {
                    TextBlock t = (TextBlock)obj;
                    t.ContextMenu = null;
                    DependencyVertex v = (DependencyVertex)t.DataContext;
                    _selectedVertex = v;
                    return true;
                }
                catch (InvalidCastException)
                {
                    return false;
                }
            }
            //else if(obj is ScrollViewer)
            //{
            //    ScrollViewer p = obj as ScrollViewer;
            //    p.ContextMenu = cxMenu;
            //    return true;
            //}
            else if (obj is DataGridColumnHeader)
            {
                DataGridColumnHeader dgColHeader = obj as DataGridColumnHeader;
                dgColHeader.ContextMenu = cxMenu;
                return true;
            }
            return false;
        }
        /*void GridViewColumnHeaderClickedHandler(object sender,
                                            RoutedEventArgs e)
        {
            GridViewColumnHeader headerClicked =
                  e.OriginalSource as GridViewColumnHeader;
            ListSortDirection direction;

            if (headerClicked != null)
            {
                if (headerClicked.Role != GridViewColumnHeaderRole.Padding)
                {
                    if (headerClicked != _lastHeaderClicked)
                    {
                        direction = ListSortDirection.Ascending;
                    }
                    else
                    {
                        if (_lastDirection == ListSortDirection.Ascending)
                        {
                            direction = ListSortDirection.Descending;
                        }
                        else
                        {
                            direction = ListSortDirection.Ascending;
                        }
                    }

                    string header = headerClicked.Column.Header as string;
                    Sort(header, direction);

                    if (direction == ListSortDirection.Ascending)
                    {
                        headerClicked.Column.HeaderTemplate =
                          Resources["HeaderTemplateArrowUp"] as DataTemplate;
                    }
                    else
                    {
                        headerClicked.Column.HeaderTemplate =
                          Resources["HeaderTemplateArrowDown"] as DataTemplate;
                    }

                    // Remove arrow from previously sorted header 
                    if (_lastHeaderClicked != null && _lastHeaderClicked != headerClicked)
                    {
                        _lastHeaderClicked.Column.HeaderTemplate = null;
                    }


                    _lastHeaderClicked = headerClicked;
                    _lastDirection = direction;
                }
            }
        }
        private void Sort(string sortBy, ListSortDirection direction)
        {
            ICollectionView dataView =
              CollectionViewSource.GetDefaultView(dataView.ItemsSource);

            dataView.SortDescriptions.Clear();
            SortDescription sd = new SortDescription(sortBy, direction);
            dataView.SortDescriptions.Add(sd);
            dataView.Refresh();
        }*/

        private void NotifyPropertyChanged(String info)
        {
            if (PropertyChanged != null)
            {
                Tracer.WriteTrace("PropertyChanged: {0}", info);
                PropertyChanged(this, new PropertyChangedEventArgs(info));
            }
        }

        private void MenuItem_Node_ShowCode(object sender, RoutedEventArgs e)
        {
            
            MenuItem m;
            DependencyVertex data;
            try
            {
                Tracer.WriteTrace("remove node: {0}", e.OriginalSource.ToString());
                m = (MenuItem)e.OriginalSource;
                data = (DependencyVertex)m.DataContext;
                if (data.GetRefObject().GetType() == typeof(SourceFile))
                    GoToSource(data.Name,true);
                else
                    GoToSource(data.FullName,false);

                //if (data != null)
                //{
                //    CodeViewer c = new CodeViewer(data.FullName, data.GetCode());
                //    c.Show();
                //}
            }
            catch (InvalidCastException)
            {
                if (_selectedVertex != null)
                {
                    if (_selectedVertex.GetRefObject().GetType() == typeof(SourceFile))
                        GoToSource(_selectedVertex.Name,true);
                    else
                        GoToSource(_selectedVertex.FullName,false);
                }
                return;
            }
            
        }
        private void GoToSource(string name,bool source)
        {
            if (name.Contains("GLOBALNAMESPACE"))
                name = name.Substring(name.LastIndexOf('.') + 1);
            IVsLibrary2 lib;
            ivs.FindLibrary(csLibGuid, out lib);
            VSOBSEARCHCRITERIA2[] sob = new[]
                    {
                        new VSOBSEARCHCRITERIA2
                            {
                                eSrchType = VSOBSEARCHTYPE.SO_SUBSTRING,
                                grfOptions = (uint) _VSOBSEARCHOPTIONS.VSOBSO_NONE,
                                szName = name
                            }
                    };
            if (source) 
                lib.GetList2((uint)_LIB_LISTTYPE.LLT_PHYSICALCONTAINERS, (uint)_LIB_LISTFLAGS.LLF_USESEARCHFILTER, sob, out  objlist);
            else
                lib.GetList2((uint)_LIB_LISTTYPE.LLT_CLASSES, (uint)_LIB_LISTFLAGS.LLF_USESEARCHFILTER, sob, out  objlist);
            //uint pCount;
            //if (objlist.GetItemCount(out pCount) == 1)
            //{
            //    int pfOK;
            //    objlist.CanGoToSource(pCount, VSOBJGOTOSRCTYPE.GS_DEFINITION, out pfOK);
            //    if (pfOK == 1)
            //    {
            //        objlist.GoToSource(pCount, VSOBJGOTOSRCTYPE.GS_DEFINITION);
            //    }
            //}
            //else
                objlist.GoToSource(0, VSOBJGOTOSRCTYPE.GS_DEFINITION);
        }
        private void MenuItem_Node_ShowEdges(object sender, RoutedEventArgs e)
        {
            MenuItem m;
            DependencyVertex data;
            try
            {
                Tracer.WriteTrace("remove node: {0}", e.OriginalSource.ToString());
                m = (MenuItem)e.OriginalSource;
                data = (DependencyVertex)m.DataContext;
                if (data != null)
                {
                    string edges = model.GetEdgeInformation(data);
                    if (edges != null)
                    {
                        CodeViewer c = new CodeViewer(data.FullName, edges);
                        c.Show();
                    }
                    else
                        MessageBox.Show("No edges connected to node!");
                }
            }
            catch (InvalidCastException)
            {
                if(_selectedVertex != null)
                {
                    string edges = model.GetEdgeInformation(_selectedVertex);
                    if (edges != null)
                    {
                        CodeViewer c = new CodeViewer(_selectedVertex.FullName, edges);
                        c.Show();
                    }
                    else
                        MessageBox.Show("No edges connected to node!");
                }
                return;
            }

        }
        internal void ShowFitness()
        {
            if(CheckInitialization())
            {
                Cursor = Cursors.Wait;
                SimulationResultViewer v = new SimulationResultViewer(model.GetFitness());
                v.Show();
                Cursor = Cursors.Arrow;
            }
        }
    }
    
}
