﻿#region License
/*
Copyright 2014 Christian Thurmann-Nielsen

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media;
using QuickGraph;
using GraphSharp;
using GraphSharp.Controls;
using java.util;
using GraphController.Utility;

namespace GraphController.Common
{
    public class DependencyVertex : DependencyObject, INotifyPropertyChanged, IEquatable<DependencyVertex>
    {
        public static readonly DependencyProperty NamespaceProperty =
            DependencyProperty.Register("Namespace", typeof(string), typeof(DependencyVertex));
        public static readonly DependencyProperty NameProperty =
            DependencyProperty.Register("Name", typeof(string), typeof(DependencyVertex));
        private String _fullName;
        private Set _referenceList;
        private List<GraphReference> _references;
        private VertexType _type;
        private VertexState _state;
        private Brush _background, _border;
        private ReferenceObject _refObject;
        private bool _isHighlighted;

        public DependencyVertex(string n, string ns)
            : this()
        {
            Name = n;
            Namespace = ns;
        }
        public DependencyVertex(string f)
        {
            Namespace = f.Substring(0, f.LastIndexOf('.'));
            Name = f.Substring((f.LastIndexOf('.') + 1), (f.Length - f.LastIndexOf('.') - 1));
        }
        public DependencyVertex()
        {
            _state = VertexState.OUTOFCYCLE;
            _references = new List<GraphReference>();
            _referenceList = new HashSet();
            _type = VertexType.INTERNAL;
        }
        public DependencyVertex(ReferenceObject c)
            : this()
        {
            _refObject = c;
            Name = c.Name;
            if (c.GetType() == typeof(Class))
                Namespace = ((Class)c).Namespace.Name;
            else
                Namespace = c.FullName;
        }
        public string Name
        {
            get { return (string)GetValue(NameProperty); }
            set { SetValue(NameProperty, value); }
        }
        public string Namespace
        {
            get { return (string)GetValue(NamespaceProperty); }
            set { SetValue(NamespaceProperty, value); _fullName = Namespace + "." + Name; }
        }
        public string ToolTip
        {
            get
            {
                if (_refObject.GetType() == typeof(Class))
                    return "Namespace: " + Namespace;
                else
                    return Namespace;
            }
        }
        public String FullName
        {
            get { return _refObject.FullName; }
        }
        public String Project
        {
            get { return _refObject.Project; }
        }
        public int FanOut
        {
            get { return _refObject.FanOut; }
        }
        public int FanIn
        {
            get { return _refObject.FanIn; }
        }
        public int MethodCount
        {
            get { return _refObject.MethodCount; }
        }
        public int StaticMethodCount
        {
            get { return _refObject.StaticMethods; }
        }
        public int MemberFields
        {
            get { return _refObject.MemberFields; }
        }
        public int Fields
        {
            get { return _refObject.Fields; }
        }
        public int Properties
        {
            get { return _refObject.Properties; }
        }
        public int Constructors
        {
            get { return _refObject.Constructors; }
        }
        public int Reachability
        {
            get { return _refObject.Reachability; }
            set { _refObject.Reachability = value; NotifyPropertyChanged("Reachability"); }
        }
        public int DependsOnSCC
        {
            get { return _refObject.DependsOnSCC; }
            set { _refObject.DependsOnSCC = value; NotifyPropertyChanged("DependsOnSCC"); }
        }
        public int SccCount
        {
            get { return _refObject.SccCount; }
            set { _refObject.SccCount = value; NotifyPropertyChanged("SccCount"); }
        }
        public int StaticFields
        {
            get { return _refObject.StaticFields; }
        }
        public int PublicStaticFields
        {
            get { return _refObject.PublicStaticFields; }
        }
        public int PrivateFields
        {
            get { return _refObject.PrivateFields; }
        }
        public bool AreAllFieldsStatic
        {
            get { return _refObject.AreAllFieldsStatic; }
        }
        public bool AreAllMethodsStatic
        {
            get { return _refObject.AreAllMethodsStatic; }
        }
        public bool IsEnum
        {
            get { return _refObject.IsEnum; }
        }
        public bool IsInterface
        {
            get { return _refObject.IsInterface; }
        }
        public bool IsPublished
        {
            get { return _refObject.IsPublished; }
        }
        public bool IsMethodsStatic
        {
            get { return _refObject.AreAllMethodsStatic; }
        }
        public bool AreFieldsStatic
        {
            get { return _refObject.AreAllFieldsStatic; }
        }
        public bool ProjectDependency
        {
            get { return _refObject.ProjectDependency; }
        }
        public VertexType Type
        {
            get { return _type; }
            set { _type = value; _border = getColor(); }
        }
        public Brush BorderBrush
        {
            get { return _border = getColor(); }
            set { _border = getColor(); NotifyPropertyChanged("Border"); }
        }
        public Brush Background
        {
            get { return _background = getBackground(); }
            set { _background = value; NotifyPropertyChanged("Background"); }
        }
        public ReferenceObject GetRefObject()
        {
            return _refObject;
        }
        public bool IsHighlighted
        {
            get { return _isHighlighted; }
            set { _isHighlighted = value; _background = getBackground(); NotifyPropertyChanged("Background"); }
        }
        public VertexState State
        {
            get { return _state; }
            set { _state = value; _background = getBackground(); NotifyPropertyChanged("Background"); }
        }
        public List<GraphReference> References
        {
            get { return _references; }
        }
        private bool ContainsReference(DependencyVertex v)
        {
            foreach (GraphReference r in _references)
            {
                if (r.Target == v)
                {
                    return true;
                }
            }
            return false;
        }
        public void AddReference(DependencyEdge e)
        {
            if (!this.ContainsReference(e.Target))
            {
                _references.Add(new GraphReference(e, e.Types));
                _referenceList.add(e.Target.FullName);
                _refObject.IncrFanOut();
                NotifyPropertyChanged("FanOut");
                e.Target.AddedAsReference();
            }
        }
        public void RemoveReference(DependencyEdge e)
        {
            if (this.ContainsReference(e.Target))
            {
                _references.RemoveAll(n => n.Target == e.Target);
                _referenceList.remove(e.Target.FullName);
                _refObject.DecrFanOut();
                NotifyPropertyChanged("FanOut");
                e.Target.RemovedAsReference();
            }
        }
        public void AddedAsReference()
        {
            _refObject.IncrFanIn();
            NotifyPropertyChanged("FanIn");
        }
        public void RemovedAsReference()
        {
            _refObject.DecrFanIn();
            NotifyPropertyChanged("FanIn");
        }
        private Brush getColor()
        {
            switch (_type)
            {
                case VertexType.EXTERNAL:
                    return Brushes.Gray;
                case VertexType.INTERNAL:
                    return Brushes.Black;
                case VertexType.PRIMITIVE:
                    return Brushes.White;
                case VertexType.UNKNOWN:
                    return Brushes.Gray;
                default:
                    return Brushes.White;
            }
        }
        private Brush getBackground()
        {
            if (_isHighlighted)
                return Brushes.Indigo;
            else
            {
                switch (_state)
                {
                    case VertexState.INCYCLE:
                        return Brushes.Orange;
                    case VertexState.OUTOFCYCLE:
                        return Brushes.White;
                    case VertexState.INDEFECTCYCLE:
                        return Brushes.Red;
                    default:
                        return Brushes.White;
                }
            }
            
        }
        
        public event PropertyChangedEventHandler PropertyChanged;
        public void NotifyPropertyChanged(string info)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new System.ComponentModel.PropertyChangedEventArgs(info));
        }


        public bool equals(object obj)
        {
            throw new NotImplementedException();
        }

        public bool Equals(DependencyVertex other)
        {
            if (_fullName == other._fullName)
                return true;
            else
                return false;
        }
        //public override int GetHashCode()
        //{
        //    return _fullName.GetHashCode();
        //}
    }
    public class GraphReference
    {
        DependencyVertex _source, _target;
        DependencyEdge _edge;
        List<ReferenceType> _types;
        private bool _isMethodStatic, _isMemberStatic, _isPublished;
        private String _tooltip;


        public GraphReference(DependencyEdge e)
        {
            _edge = e;
            _source = e.Source;
            _target = e.Target;
            _types = new List<ReferenceType>();
            _isMethodStatic = false;
            _isMemberStatic = false;
            _isPublished = false;
        }
        public GraphReference(DependencyEdge e, ReferenceType type)
            : this(e)
        {
            _types.Add(type);
            GenerateTooltip();
        }
        public GraphReference(DependencyEdge e, List<ReferenceType> types)
            : this(e)
        {
            _types = types;
            GenerateTooltip();
        }

        public DependencyVertex Source
        {
            get { return _source; }
        }
        public DependencyVertex Target
        {
            get { return _target; }
        }
        public List<ReferenceType> Types
        {
            get { return _types; }
            set { _types = value; }
        }
        public String Tooltip
        {
            get { return _tooltip; }
            set { _tooltip = value; }
        }
        public bool ContainsType(ReferenceType t)
        {
            return _types.Contains(t);
        }
        public bool IsMemberStatic
        {
            get { return _isMemberStatic; }
            set { _isMemberStatic = value; }
        }

        public bool IsMethodStatic
        {
            get { return _isMethodStatic; }
            set { _isMethodStatic = value; }
        }
        public bool IsPublished
        {
            get { return _isPublished; }
            set { _isPublished = value; }
        }
        private void GenerateTooltip()
        {
            string s = "";

            if (_types.Count <= 0)
                return;

            if(_types.Count == 1)
            {
                _tooltip = GraphData.TypeString(_types.First());
                return;
            }
            for (int i = 0; i < _types.Count - 1; i++)
            {
                if (_types[i] == ReferenceType.USES)
                    continue;
                if (!s.Contains(GraphData.TypeString(_types[i])))
                    s += GraphData.TypeString(_types[i]) + ", ";
            }
            s += GraphData.TypeString(_types.Last());            

            _tooltip = s;
        }
        public bool CanExtractInterface()
        {
            if (!_types.Contains(ReferenceType.METHODPARAMETER) || _types.Contains(ReferenceType.METHODRETURNTYPE))
                return true;
            else
                return false;
        }
        public void AddReferenceType(ReferenceType t)
        {
            if(!_types.Contains(t))
            {
                _types.Add(t);
                GenerateTooltip();
                _edge.NotifyPropertyChanged("TypeString");
            }
        }
        public bool RemoveReference(ReferenceType t)
        {
            if (_types.Contains(t))
            {

                if (_types.Remove(t))
                {
                    GenerateTooltip();
                    _edge.NotifyPropertyChanged("TypeString");
                    return true;
                }
                else
                    return false;
            }
            else
                return false;
        }
        
    }
    public class CycleData
    {
        private ReferenceObject _mainClass;
        private int _size;
        private List<ReferenceObject> _vertices;
        private Set cycle;

        public CycleData(List<ReferenceObject> l, ReferenceObject mainClass)
        {
            _vertices = l;
            _mainClass = mainClass;
            _size = _vertices.Count;
        }
        public CycleData(List<ReferenceObject> l, ReferenceObject mainClass, Set s)
        {
            _vertices = l;
            _mainClass = mainClass;
            _size = _vertices.Count;
            cycle = s;
        }

        public ReferenceObject MainClass
        {
            get { return _mainClass; }
            set { _mainClass = value; }
        }

        public int Size
        {
            get { return _size; }
            set { _size = value; }
        }
        public List<ReferenceObject> Vertices
        {
            get { return _vertices; }
            set { _vertices = value; }
        }
        public Set CycleSet
        {
            get { return cycle; }
        }
        public override string ToString()
        {
            return _mainClass.FullName + ": " + _size;
        } 
        public void PrindCycleSet()
        {
            HashSet s = (HashSet)cycle;
            foreach (object o in s)
            {
                Tracer.WriteTrace("{0}", o);
            }
        }
    }
    public class EdgeTag
    {
    }
    public class DependencyEdge : TaggedEdge<DependencyVertex, object>,IEdgeCommand,INotifyPropertyChanged, IEquatable<DependencyEdge>
    {
        private GraphReference _ref;
        private Brush _edgeColor;
        private bool _isCyclic;
        private DependencyEdge _cyclicSibling;

        public DependencyEdge(DependencyVertex fromvertex, DependencyVertex tovertex, ReferenceType type)
            : base(fromvertex, tovertex, EdgeTypes.General)
        {
            _ref = new GraphReference(this,type);
            _isCyclic = false;
            _edgeColor = GetColor();
        }
        public DependencyEdge(DependencyVertex fromvertex, DependencyVertex tovertex, List<ReferenceType> types)
            : base(fromvertex, tovertex, EdgeTypes.General)
        {
            _ref = new GraphReference(this, types);
            _isCyclic = false;
            _edgeColor = GetColor();
        }
        public DependencyEdge(DependencyVertex s,DependencyVertex t, DependencyEdge e)
            : base(s,t,EdgeTypes.General)
        {
            _ref = new GraphReference(e);
            foreach(ReferenceType type in e.Reference.Types)
            {
                AddReferenceType(type);
            }
        }
        public string Type
        {
            get { return _ref.Types.ToString(); }
        }
        public GraphReference Reference
        {
            get { return _ref; }
        }
        public bool IsCyclic
        {
            get { return _isCyclic; }
            set { _isCyclic = value; _edgeColor = GetColor(); NotifyPropertyChanged("EdgeColor"); }
        }
        public Brush EdgeColor
        {
            get { return _edgeColor; }
            set { _edgeColor = value; NotifyPropertyChanged("EdgeColor"); }
        }
        public List<ReferenceType> Types
        {
            get { return _ref.Types; }
        }
        public DependencyEdge CyclicSibling
        {
            get { return _cyclicSibling; }
            set { _cyclicSibling = value; }
        }
        public string TypeString
        {
            get
            {
                return _ref.Tooltip;
            }
            
        }
        public void AddReferenceType(ReferenceType t)
        {
            if(!_ref.ContainsType(t))
                _ref.AddReferenceType(t);
        }
        
        private Brush GetColor()
        {
            
            if (_isCyclic)
                return Brushes.Red;
            else
                return Brushes.Green;
        }
        public void OnDoubleClick()
        {
            Tracer.WriteTrace("Edge, RightClick");
        }

        public void OnRightClick()
        {
            Tracer.WriteTrace("Edge, DoubleClick");
        }
        public string GetEdgeInfo()
        {
            StringBuilder sb = new StringBuilder();

            if (_isCyclic && _cyclicSibling != null)
            {
                sb.Append("Edge is cyclic!" + System.Environment.NewLine + System.Environment.NewLine);
                sb.Append("First Edge:" + System.Environment.NewLine);
                sb.Append(this.ToString() + System.Environment.NewLine + System.Environment.NewLine);
                sb.Append("Second Edge: " + System.Environment.NewLine);
                sb.Append(_cyclicSibling.ToString());
            }
            else
            {
                sb.Append("Edge is acyclic." + System.Environment.NewLine);
                sb.Append(this.ToString());
            }


            return sb.ToString();
        }
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();

            sb.Append("     Source: " + Source.FullName + System.Environment.NewLine);
            sb.Append("     Target: " + Target.FullName + System.Environment.NewLine);

            if (_ref.Types.Count > 1)
            {
                sb.Append("     Types: " +  GraphData.TypeString(_ref.Types.First()));
                for (int i = 1; i < _ref.Types.Count - 1; i++)
                {
                    if(!sb.ToString().Contains( GraphData.TypeString(_ref.Types[i])))
                        sb.Append(", " + GraphData.TypeString(_ref.Types[i]) );
                }
                if (!sb.ToString().Contains(GraphData.TypeString(_ref.Types.Last())))
                    sb.Append(", " + GraphData.TypeString(_ref.Types.Last()));
            }
            else if (_ref.Types.Count > 0)
            {
                sb.Append("     Type: " + GraphData.TypeString(_ref.Types.First()));
            }
            else
                sb.Append("     No Types!");


            return sb.ToString();
        }

        public event PropertyChangedEventHandler PropertyChanged;
        public void NotifyPropertyChanged(string info)
        {
            PropertyChangedEventHandler handler = PropertyChanged;
            if (handler != null)
                handler(this, new System.ComponentModel.PropertyChangedEventArgs(info));
        }

        public bool Equals(DependencyEdge other)
        {
            if (Source.FullName == other.Source.FullName && Target.FullName == other.Target.FullName)
                return true;
            else
                return false;
        }
    }
    public interface IEdgeCommand
    {
        void OnDoubleClick();
        void OnRightClick();
    }

    public enum VertexType
    {
        EXTERNAL, INTERNAL, UNKNOWN, PRIMITIVE
    }
    public enum VertexState
    {
        INCYCLE, OUTOFCYCLE, INDEFECTCYCLE
    }
    public enum ReferenceType
    {
        MEMBERREFERENCE, STATICMEMBERREFERENCE, METHODINVOCATION, STATICMETHODINVOCATION, METHODRETURNTYPE, METHODRETURNTYPEPUB, METHODPARAMETER, METHODPARAMETERPUB, METHODBODY, BASE, INTERFACE, MEMBERFIELD, MEMBERPROPERTY, STATICFIELD, EXCEPTION, CONSTRUCTORINVOCATION, USES, IMPLEMENTS, ADDEDINGUI, UNKNOWN, NODE, DEPENDENCY
    }
    public static class GraphData
    {
            public static List<ReferenceType> GetAllTypes()
            {
                List<ReferenceType> l = new List<ReferenceType>();
                l.Add(ReferenceType.METHODINVOCATION);
                l.Add(ReferenceType.STATICMETHODINVOCATION);
                l.Add(ReferenceType.METHODRETURNTYPE);
                l.Add(ReferenceType.METHODPARAMETER);
                l.Add(ReferenceType.MEMBERFIELD);
                l.Add(ReferenceType.STATICFIELD);
                l.Add(ReferenceType.BASE);
                l.Add(ReferenceType.INTERFACE);
                l.Add(ReferenceType.MEMBERPROPERTY);
                l.Add(ReferenceType.CONSTRUCTORINVOCATION);
                l.Add(ReferenceType.MEMBERREFERENCE);
                l.Add(ReferenceType.STATICMEMBERREFERENCE);

                return l;
            }
        public static string TypeString(ReferenceType t)
        {
            switch (t)
            {
                case ReferenceType.BASE:
                    return "BASE";
                case ReferenceType.INTERFACE:
                    return "INTERFACE";
                case ReferenceType.MEMBERFIELD:
                    return "MEMBERFIELD";
                case ReferenceType.MEMBERREFERENCE:
                    return "MEMBERREFERENCE";
                case ReferenceType.STATICMEMBERREFERENCE:
                    return "STATICMEMBERREFERENCE";
                case ReferenceType.METHODBODY:
                    return "METHODBODY";
                case ReferenceType.METHODPARAMETER:
                    return "METHODPARAMETER";
                case ReferenceType.METHODPARAMETERPUB:
                    return "METHODPARAMETER";
                case ReferenceType.METHODRETURNTYPE:
                    return "METHODRETURNTYPE";
                case ReferenceType.METHODRETURNTYPEPUB:
                    return "METHODRETURNTYPE";
                case ReferenceType.STATICFIELD:
                    return "STATICFIELD";
                case ReferenceType.CONSTRUCTORINVOCATION:
                    return "CONSTRUCTORINVOCATION";
                case ReferenceType.METHODINVOCATION:
                    return "METHODINVOCATION";
                case ReferenceType.STATICMETHODINVOCATION:
                    return "STATICMETHODINVOCATION";
                case ReferenceType.MEMBERPROPERTY:
                    return "MEMBERPROPERTY";
                default:
                    return "UNKOWN";
            } 
        }
    }
}
