﻿#region License
/*
Copyright 2014 Christian Thurmann-Nielsen

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using java.util;
using melton.hayden.jepends.analysis;
using tosin.oyetoyan.research.support;
using tosin.oyetoyan.research.main.utility;
using GraphController.Utility;
using GraphController.GUI;

namespace GraphController.Common
{
    public class GraphLogic
    {
        private MapModel _model, _previous;
        private DependencyGraph _graph;
        private Map _originalMap;
        private Map _reachability, _dependOnSCC;
        private Analyser _analyser;
        private List _sccs, _oldSccs;
        private List<CycleData> _cycleData;
        private List _table1Result, _table2Result, _decCyclesResult, _modelSimResult; 
        private ExtractInterfaceSupportTable1 _table1;
        private DecisionSupportTable2 _table2;
        private ModelSimulator _modelSim;
        private DecisionSupportCycles _decCycles;
        private int _topKFanin, _topKCrss, _topKScc, _numOfRefactoring;
        private double _reductionRate;
        private bool _moveAssociated;
        private string _mainClass;
        private string sortingPrio;



        public GraphLogic(MapModel model,DependencyGraph graph)
        {
            _model = model;
            _graph = graph;
            _analyser = new Analyser();
            _cycleData = new List<CycleData>();
            _topKCrss = 40;
            _topKFanin = 40;
            _topKScc = 25;
            sortingPrio = ExtractInterfaceSupportTable1.CRSS;
            _moveAssociated = false;
        }
        public GraphLogic(MapModel m,DependencyGraph g,GraphLogic l)
            : this(m,g)
        {
            _topKCrss = l._topKCrss;
            _topKFanin = l._topKFanin;
            _topKScc = l._topKScc;
            _mainClass = l._mainClass;
        }

        #region Encapsulation Methods
        public int TopKScc
        {
            get { return _topKScc; }
            set { _topKScc = value; }
        }

        public int TopKCrss
        {
            get { return _topKCrss; }
            set { _topKCrss = value; }
        }

        public int TopKFanin
        {
            get { return _topKFanin; }
            set { _topKFanin = value; }
        }
        public string MainClass
        {
            get { return _mainClass; }
            set { _mainClass = value; }
        }
        public int NumOfRefactoring
        {
            get { return _numOfRefactoring; }
            set { _numOfRefactoring = value; }
        }
        public double ReductionRate
        {
            get { return _reductionRate; }
            set { _reductionRate = value; }
        }
        public MapModel Model
        {
            get { return _model; }
            set { _model = new MapModel(value); }
        }
        public DependencyGraph Graph
        {
            get { return _graph; }
            set { _graph = value; }
        }
        public List<CycleData> CycleData
        {
            get { return _cycleData; }
            set { _cycleData = value; }
        }
        #endregion
        public void Analyse()
        {
            Analyse(false);
        }
        public void AnalyseAndUpdateView()
        {
            Analyse(true);
        }
        private void Analyse(bool flag)
        {
            _analyser = new Analyser();
            _cycleData = new List<CycleData>();
            Map m = _model.AllDependencies;
            _previous = new MapModel(_model);
            _analyser.analyse(m);
            _oldSccs = _sccs;
            _sccs = _analyser.getSccs();
            _dependOnSCC = _analyser.getDependOnSCC();
            CalculateReachability(flag);
            CalculateDependOnSCC();
        }
        private void CalculateReachability(bool flag)
        {
            Map m = _model.AllDependencies;
            melton.hayden.jepends.algorithm.ReachabilityMapGenerator reachability = new melton.hayden.jepends.algorithm.ReachabilityMapGenerator(m);
            _reachability = reachability.getMap();

            object[] keySet = _reachability.keySet().toArray();
            object[] values = _reachability.values().toArray();

            for (int i = 0; i < keySet.Length; i++)
            {
                HashSet reach = (HashSet)values[i];
                string s = (string)keySet[i];
                var l = _graph.Vertices.Where(n => n.FullName == s);
                foreach (DependencyVertex v in l)
                {
                    v.Reachability = reach.size();
                }
            }
            if (flag)
                ColorSccs();
            else
                CalculateCycleData();
        }
        public void Rollback()
        {
            _model = _previous;
        }
        public void SetSortingPriority(string s)
        {
            switch(s)
            {
                case "SCC":
                    sortingPrio = ExtractInterfaceSupportTable1.SCC;
                    break;
                case "CRSS":
                    sortingPrio = ExtractInterfaceSupportTable1.CRSS;
                    break;
                case "STATIC":
                    sortingPrio = ExtractInterfaceSupportTable1.STATIC;
                    break;
                case "FANIN":
                    sortingPrio = ExtractInterfaceSupportTable1.FANIN;
                    break;
                default:
                    sortingPrio = ExtractInterfaceSupportTable1.CRSS;
                    break;
            }
        }
        public void DecisionTable1()
        {
            //Table1 is needed for SystemRefactoring
            
            _table1 = new ExtractInterfaceSupportTable1(_model.AllDependencies, _model.PublishedTypes, _sccs, _model.MethodTypes, _model.ClassTypes, sortingPrio);
            //sccList from analyser: List<Set<string>>
            _table1.run();
                
        }
        public void DecisionTable2(bool flag)
        {
            if (flag)
            {
                DecisionTable1();
            }

            _table2 = new DecisionSupportTable2(_table1.getComps_EI(), _model.MethodSize, _model.AllDependencies);
            //_table2.setTopK(_topKScc);
            _table2.setCRSSReducedRate(_reductionRate);
            _table2.setNonStaticFieldMap(_model.Memberfields);
            _table2.setNonStaticMethodMap(_model.MethodInvocations);
            _table2.setInterfaceMap(_model.Interfacetypes);
            _table2.setStaticFieldMap(_model.StaticMemberUsage);
            _table2.setStaticMethodMap(_model.StaticMethodInvocations);
            _table2.setSuperMap(_model.SuperTypes);
            _table2.setConstructorInvocationMap(_model.Constructorinvocations);
            _table2.run();
        }
        
        public List GetSingleComponentCandidateData()
        {
            //table1.update() if it has been run previously
            //Table 2 is needed for SigleComponent and CycleRefactoring
            DecisionTable2(true);
            _table2.processCandidates(TopKCrss,DecisionSupportTable2.CRSS);
            
            _table2Result = _table2.getCandidateDataCSV();

            Tracer.WriteTrace("Single component candidates: {0}",_table2Result.size());
            return _table2Result; //Set<string> where the string is commaseparated with "source,target,crss,fanout,scc,reductionrate,decision" 
             
            // Decision is either SI = StaticFieldInling, DI = ExtractInterface or InterfaceInjection, SR = SingletonRegistry, NONE = NoDecision
        }
        public List GetCycleCandidates(Set scc, bool flag)
        {
            //table1.updateTable() if it has been run previously or map has been updated
            DecisionTable2(true);
            _moveAssociated = flag;
            _decCycles = new DecisionSupportCycles(_table1.getComps_EI(), _model.AllDependencies, _table2.getDsTableSC_lookUp());
            //_decCycles = new DecisionSupportCycles(_table1.getComps_EI(), _model.AllDependencies, _model.SuperTypes, _table2.getDsTableSC_lookUp()); // SuperTypes will be removed, just return null for now
            
            string sorting = DecisionSupportCycles.FANIN_TARGET;

            _decCycles.run(scc, sorting); //sortingParameter could be either "reducedby" or "fanin"
            //scc is from the analyser choose one cycle from the list

            return _decCycles.getCandidateDataCSV();//_decCycles.getCandidateCSV(); //Set<string> where the string is commaseparated with "from,to,crss,scc,reductionrate,decision" 

            //modelSim.simulateOthers(List<string> source,List<string> targets,List<string> decision)
        }
        public List GetSystemCandidateData(double rate)
        {
            //table1.update() if it has been run previously
            //Table 2 is needed for SigleComponent and CycleRefactoring
            _table1 = new ExtractInterfaceSupportTable1(_model.AllDependencies, _model.PublishedTypes, _sccs, _model.MethodTypes, _model.ClassTypes, sortingPrio);
            //sccList from analyser: List<Set<string>>
            _table1.run();
            _table1.processCandidates(_topKCrss, _topKScc, _topKFanin, rate); //Only for system refactoring!
            


            _table1Result = _table1.getCandidateDataCSV();
            Tracer.WriteTrace("System candidates: {0}", _table1Result.size());
            return _table1Result;
            // Decision is either SI = StaticFieldInling, DI = ExtractInterface or InterfaceInjection, SR = SingletonRegistry, NONE = NoDecision
        }
        private void InitModelSimulator()
        {
            _modelSim = new ModelSimulator(_model.AllDependencies, _mainClass, sortingPrio);
            _modelSim.setTopKcrss(_topKCrss);
            _modelSim.setTopKscc(_topKScc);
            _modelSim.setTopKfanin(_topKFanin);
            _modelSim.setReducedRate(_reductionRate);
            _modelSim.setMethodTypeMap(_model.MethodTypes);
            _modelSim.setPublishedTypeMap(_model.PublishedTypes);
            _modelSim.setClassTypeMap(_model.ClassTypes);
            _modelSim.setStaticFieldMap(_model.StaticMemberUsage);
            _modelSim.setStaticMethodMap(_model.StaticMethodInvocations);
            _modelSim.setConstructorInvocationMap(_model.Constructorinvocations);
            _modelSim.setNonStaticMethodMap(_model.MethodInvocations);
            _modelSim.setSupersMap(_model.SuperTypes);
            _modelSim.setSupportTableSC_lookUp(_table1.getComps_EI());
            _modelSim.setMoveAssociatedDependencies(_moveAssociated);
        }
       
        private string TranslateDecisionEnum(RefactoringStrategy s)
        {
            switch(s)
            {
                case RefactoringStrategy.EXTRACTINTERFACE:
                    return DecisionSupportTable2.Store.EXTRACTINTERFACE;
                case RefactoringStrategy.PROTOTYPEREGISTRY:
                    return DecisionSupportTable2.Store.PROTOTYPEREGISTRY;
                case RefactoringStrategy.SINLGETONREGISTRY:
                    return DecisionSupportTable2.Store.SINGLETONREGISTRY;
                case RefactoringStrategy.STATICFINALFIELDCOPY:
                    return DecisionSupportTable2.Store.STATICFINALFIELDCOPY;
                case RefactoringStrategy.STATICMETHODINLINING:
                    return DecisionSupportTable2.Store.STATICMETHODINLINING;
            }
            return "";
        }
        private void PrintList(List l,string name)
        {
            Tracer.WriteTrace("Printing List: {0}",name);
            if (l == null)
                Tracer.WriteTrace("\tNULL");
            else
            {
                for (int i = 0; i < l.size(); i++)
                {
                    Tracer.WriteTrace("\tindex {0}: {1}", i, l.get(i));
                }
            }
        }
        public SimulationResult SimulateCycleRefactoring(List<CycleRefactoringData> list)
        {
            //_modelSim = new ModelSimulator(_model.AllDependencies, _mainClass, sortingPrio);
            //_modelSim.setTopKcrss(_topKCrss);
            //_modelSim.setTopKscc(_topKScc);
            //_modelSim.setTopKfanin(_topKFanin);
            //_modelSim.setReducedRate(_reductionRate);
            //_modelSim.setMethodTypeMap(_model.MethodTypes);
            //_modelSim.setPublishedTypeMap(_model.PublishedTypes);
            //_modelSim.setClassTypeMap(_model.ClassTypes);
            //_modelSim.setStaticFieldMap(_model.StaticMemberUsage);
            //_modelSim.setStaticMethodMap(_model.StaticMethodInvocations);
            //_modelSim.setConstructorInvocationMap(_model.Constructorinvocations);
            //_modelSim.setNonStaticMethodMap(_model.MethodInvocations);
            //_modelSim.setSupersMap(_model.SuperTypes);
            //_modelSim.setSupportTableSC_lookUp(_table1.getComps_EI());
            InitModelSimulator();

            PrintList(_table2.getDecisions(), "Table 2 Decisions");
            if (list.Count > 0)
            {
                List sources = new ArrayList();
                List targets = new ArrayList();
                List decisions = new ArrayList();
                foreach (CycleRefactoringData c in list)
                {
                    sources.add(c.Source);
                    targets.add(c.Target);
                    decisions.add(c.GetOriginalDecision());
                }
                _modelSim.simulateOthers(sources, targets, decisions); 

                List l = _modelSim.getFinalCand();
                Tracer.WriteTrace("Final Candidates:\n");
                for (int i = 0; i < l.size(); i++)
                {
                    Tracer.WriteTrace("\t{0}: {1}", i, l.get(i));
                }
            }
            else
            {
                _modelSim.simulateOthers(_table2.getSources(), _table2.getTargets(), _table2.getDecisions());
            }
            return GetSimulationResult();
        }
        public SimulationResult SimulateSystemRefactoring(int num)
        {
            //table1.update() if it has been run previously
            //InitModelSimulator(false);
            //_table1.processCandidates(_topKCrss, _topKScc, _topKFanin, _reductionRate);
            //PrintList(_table1.getCandidates(), "Table1 Candidates");
            //if (list.Count > 0)
            //{
            //    List candidates = new ArrayList();
            //    foreach (SystemRefactoringData c in list)
            //    {
            //        candidates.add(c.Class);
            //    }
            //    _modelSim.simulateSystem(candidates, num, true);
            //}
            //else
            //    
            //_modelSim = new ModelSimulator(_model.AllDependencies, _mainClass, sortingPrio);
            //_modelSim.setTopKcrss(_topKCrss);
            //_modelSim.setTopKscc(_topKScc);
            //_modelSim.setTopKfanin(_topKFanin);
            //_modelSim.setReducedRate(_reductionRate);
            //_modelSim.setMethodTypeMap(_model.MethodTypes);
            //_modelSim.setPublishedTypeMap(_model.PublishedTypes);
            //_modelSim.setClassTypeMap(_model.ClassTypes);
            //_modelSim.setStaticFieldMap(_model.StaticMemberUsage);
            //_modelSim.setStaticMethodMap(_model.StaticMethodInvocations);
            //_modelSim.setConstructorInvocationMap(_model.Constructorinvocations);
            //_modelSim.setNonStaticMethodMap(_model.MethodInvocations);
            //_modelSim.setSupersMap(_model.SuperTypes);
            //_modelSim.setSupportTableSC_lookUp(_table1.getComps_EI());
            //_modelSim.setSccs(_sccs);
            InitModelSimulator();
            _modelSim.simulateSystem(_table1.getCandidates(), num, true);

            List l = _modelSim.getFinalCand();
            Tracer.WriteTrace("Final Candidates:\n");
            for (int i = 0; i < l.size(); i++ )
            {
                Tracer.WriteTrace("\t{0}: {1}", i, l.get(i));
            }

            return GetSimulationResult();

            //GetSubgraph() is renamed into GetRafactoredEdges() returns a List<RefactoredEdges> 
            //RefactoredEdges contains SOurce, Target, Relationship
        }
        public SimulationResult SimulateSingleComponentRefactoring(List<SingleComponentRefactoringData> list)
        {
            DecisionTable2(true);
            InitModelSimulator();
            if (list.Count > 0)
            {
                List sources = new ArrayList();
                List targets = new ArrayList();
                List decisions = new ArrayList();
                foreach (SingleComponentRefactoringData c in list)
                {
                    sources.add(c.Source);
                    targets.add(c.Target);
                    decisions.add(c.GetOriginalDecision());
                }
                _modelSim.simulateOthers(sources, targets, decisions);
            }
            else
            {
                _modelSim.simulateOthers(_table2.getSources(), _table2.getTargets(), _table2.getDecisions());
            }
            return GetSimulationResult();
        }
        public SimulationResult GetFitness()
        {
            SimulationResult r = new SimulationResult();
            tosin.oyetoyan.research.support.fitness.FitnessValue v = (new FitnessCalc(_model.AllDependencies)).getFitness();

            r.Fitness = v;
            
            return r;
        }
        private SimulationResult GetSimulationResult()
        {
            SimulationResult s = new SimulationResult();

            try
            {
                s.NumberOfRefactorings = _modelSim.getFinalRefactorNumber();
                s.RefactoredEdges = _modelSim.getRefactoredEdges();
                s.Fitness = _modelSim.getFitness();
                s.FinalCandidates = _modelSim.getFinalCand();
                s.RefactoringReport = _modelSim.getReportRefactors();
            }
            catch(NullReferenceException)
            {
                return null;
            }

            return s;
        }
        private void CalculateDependOnSCC()
        {
            object[] keys = _dependOnSCC.keySet().toArray();

            for(int i = 0; i < keys.Count(); i++)
            {
                string className = (string)keys[i];
                Set set =  (Set)_dependOnSCC.get(className);
                int size = set.size();
                var v = _graph.Vertices.Where(n => n.FullName == className);
                DependencyVertex vertex = v.First();
                vertex.DependsOnSCC = size;
            }
        }
        private void CalculateCycleData()
        {
            _cycleData = new List<CycleData>();
            List l = _sccs;

            for (int i = 0; i < l.size(); i++)
            {
                HashSet arr = (HashSet)l.get(i);
                int highest = 0;
                ReferenceObject mainCycleClass = null;
                List<ReferenceObject> list = new List<ReferenceObject>();

                foreach (string s in arr)
                {
                    var v = _graph.Vertices.Where(n => n.FullName == s);
                    foreach (DependencyVertex vertex in v)
                    {
                        if (vertex.Reachability > highest)
                        {
                            highest = vertex.Reachability;
                            mainCycleClass = vertex.GetRefObject();
                        }
                        list.Add(vertex.GetRefObject());
                    }

                }
                CycleData cd = new CycleData(list, mainCycleClass, arr);
                _cycleData.Add(cd);
            }
        }
        private void ColorSccs()
        {
            List l = _sccs;
            _cycleData = new List<CycleData>();

            //_graph.ColorSccs(l);
            if (_oldSccs != null)
            {
                Tracer.WriteTrace("ColorSccs: from analyser: {0}, current: {1}", l.size(), _oldSccs.size());
                for (int i = 0; i < _oldSccs.size(); i++)
                {
                    HashSet arr = (HashSet)_oldSccs.get(i);

                    foreach (string s in arr)
                    {
                        var v = _graph.Vertices.Where(n => n.FullName == s);
                        foreach (DependencyVertex vertex in v)
                        {
                            Tracer.WriteTrace("ColorSccs() -> uncoloring: {0}", vertex.FullName);
                            vertex.State = VertexState.OUTOFCYCLE;
                        }
                    }
                }
            }
            else
                Tracer.WriteTrace("ColorSccs: from analyser: {0}", l.size());

            for (int i = 0; i < l.size(); i++)
            {
                HashSet arr = (HashSet)l.get(i);
                int highest = 0;
                ReferenceObject mainCycleClass = null;
                List<ReferenceObject> list = new List<ReferenceObject>();

                foreach (string s in arr)
                {
                    var v = _graph.Vertices.Where(n => n.FullName == s);
                    foreach (DependencyVertex vertex in v)
                    {
                        Tracer.WriteTrace("ColorSccs() -> coloring: {0}", vertex.FullName);
                        vertex.State = VertexState.INCYCLE;
                        if (vertex.Reachability > highest)
                        {
                            highest = vertex.Reachability;
                            mainCycleClass = vertex.GetRefObject();
                        }
                        list.Add(vertex.GetRefObject());
                        vertex.SccCount = arr.size();
                    }

                }
                CycleData cd = new CycleData(list, mainCycleClass,arr);
                _cycleData.Add(cd);
            }

            _sccs = l;
        }
    }
}
