﻿#region License
/*
Copyright 2014 Christian Thurmann-Nielsen

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using GraphController.Utility;

namespace GraphController.Common
{
    /// <summary>
    /// Since a Namespace can be divided into several classfiles I need to keep track of which 'using' statements belong to which classfile
    /// a Namespace is represented in.
    /// </summary>
    public class Namespace
    {
        List<BaseClass> _classes;
        private string _name;
        DependencyGraph _graph;

        public Namespace(string s)
        {
            _name = s;
            _graph = new DependencyGraph();
            _classes = new List<BaseClass>();
        }
        public DependencyGraph Graph
        {
            get { return _graph; }
        }
        public string Name
        {
            get { return _name; }
        }
        public List<BaseClass> Classes
        {
            get { return _classes; }
        }
        public void AddClass(BaseClass c)
        {
            _classes.Add(c);
        }
        public bool IsClassInNamespace(string s)
        {
            foreach (BaseClass c in _classes)
            {
                if (c.Name == s)
                    return true;
            }
            return false;
        }
    }
    public class CodeReference
    {
        private ReferenceObject _source, _target;
        private SourceFile _sourcefileSource;
        private ReferenceType _type;
        private bool _isPublished, _isStatic;
        private string _memberProperty;


        public CodeReference(ReferenceObject s, ReferenceObject t, ReferenceType type)
        {
            _source = s;
            _target = t;
            _type = type;
            _isPublished = false;
            _isStatic = false;
        }
        public CodeReference(ReferenceObject s, ReferenceObject t, ReferenceType type, bool published, bool isStatic)
            : this(s,t,type)
        {
            _isPublished = published;
            _isStatic = isStatic;
        }
        /// <summary>
        /// Only used for partial classes where the sourcefile depenendency also need to be determined
        /// </summary>
        /// <param name="s"></param>
        /// <param name="t"></param>
        /// <param name="type"></param>
        /// <param name="published"></param>
        /// <param name="isStatic"></param>
        /// <param name="source"></param>
        /// <param name="target"></param>
        public CodeReference(ReferenceObject s, ReferenceObject t, ReferenceType type, bool published, bool isStatic,SourceFile source)
            : this(s, t, type, published, isStatic)
        {
            if (source == null)
                throw new NullReferenceException();
            _sourcefileSource = source;
        }
        public ReferenceObject Target
        {
            get { return _target; }
        }

        public ReferenceObject Source
        {
            get { return _source; }
        }
        public SourceFile SourceFile
        {
            get { return _sourcefileSource; }
        }
        public ReferenceType Type
        {
            get { return _type; }
            set { _type = value; }
        }
        public bool IsPublished
        {
            get { return _isPublished; }
            set { _isPublished = value; }
        }
        public bool IsStatic
        {
            get { return _isStatic; }
            set { _isStatic = value; }
        }
        public string MemberProperty
        {
            get { return _memberProperty; }
            set { _memberProperty = value; }
        }

    }
    public class Class : BaseClass
    {

        private SourceFile _sourceFile;

        public Class()
            : base() {}

        public Class(string s,string sf)
            : base(s,sf){}

        public Class(string n, Namespace ns,SourceFile c,string p)
            : base(n,ns,p)
        {
            _sourceFile = c;
        }

        public Class(Class c)
            : base(c)
        {
            _sourceFile = c._sourceFile;
        }
        public Class(string n) // The Classes that will be initialized with this constructor does not need a list of references.
            : base(n) {}
        public SourceFile SourceFile
        {
            get { return _sourceFile; }
        }
    }
    public class PartialClass : BaseClass
    {

        private List<SourceFile> _sourceFiles;

        public PartialClass(string n, Namespace ns, SourceFile c, string project)
            : base(n,ns,project)
        {
            _sourceFiles = new List<SourceFile>();
            _sourceFiles.Add(c);
        }

        public PartialClass(PartialClass c)
            : base(c)
        {
            _sourceFiles = c._sourceFiles;
        }
        public List<SourceFile> SourceFiles
        {
            get { return _sourceFiles; }
        }

        public void AddSourcefile(SourceFile s)
        {
            if (!_sourceFiles.Contains(s))
            {
                _sourceFiles.Add(s);
            }
        }
    }
    public class Member
    {
        string _type, _name;
        bool _isStatic;

        public Member(string t, string n,bool s)
        {
            _type = t;
            _name = n;
            _isStatic = s;
        }
        public string Name
        {
            get { return _name; }
        }
        public string Type
        {
            get { return _type; }
        }
        public bool IsStatic
        {
            get { return _isStatic; }
        }
    }
    public abstract class BaseClass : ReferenceObject
    {
        private Namespace _namespace;
        private Dictionary<string, string> _methodReturnTypes;
        private List<Member> _memberProperties;
        private List<Member> _memberFields;

        protected BaseClass(string n, Namespace ns,string p)
            : this()
        {
            _name = n;
            _namespace = ns;
            _fullName = ns.Name + "." + _name;
            _vertex = new DependencyVertex(this);
            _project = p;
        }
        protected BaseClass()
            : base()
        {
            _references = new List<CodeReference>();
            _methodReturnTypes = new Dictionary<string, string>();
            _memberProperties = new List<Member>();
            _memberFields = new List<Member>();
        }
        protected BaseClass(string s, string sf)
        {
            string ns = "RefactorNamespace";
            if(s.Contains('.'))
            {
                ns = s.Substring(0, s.LastIndexOf('.'));
                _name = s.Substring((s.LastIndexOf('.') + 1), (s.Length - s.LastIndexOf('.') - 1));
            }
            else
                _name = s;

            _namespace = new Namespace(ns);
            _fullName = _namespace.Name + "." + _name;
            _vertex = new DependencyVertex(this);
        }
        protected BaseClass(string n) // The Classes that will be initialized with this constructor does not need a list of references.
        {
            _name = n;
        }
        protected BaseClass(BaseClass c)
        {
            _areFieldsStatic = c._areFieldsStatic;
            _areMethodsStatic = c._areMethodsStatic;
            _constructors = c._constructors;
            _containsNonStatic = c._containsNonStatic;
            _containsStatic = c._containsStatic;
            _fanIn = 0;
            _fanOut = 0;
            _fields = c._fields;
            _fullName = c._fullName;
            _isEnum = c._isEnum;
            _isInterface = c._isInterface;
            _isPublished = c._isPublished;
            _projectDependency = c._projectDependency;
            _methods = c._methods;
            _name = c._name;
            _namespace = c._namespace;
            _parent = c._parent;
            _privateFields = c._privateFields;
            _project = c._project;
            _properties = c._properties;
            _publicStaticFields = c._publicStaticFields;
            _staticFields = c._staticFields;
            _staticMethods = c._staticMethods;
            _vertex = c._vertex;
            _references = c._references;
        }
       
        public Namespace Namespace
        {
            get { return _namespace; }
            set { _namespace = value; _fullName = _namespace.Name + "." + _name; _vertex.Namespace = _namespace.Name; }
        }
        public DependencyVertex Vertex
        {
            get { return _vertex; }
        }
        public List<CodeReference> References
        {
            get { return _references; }
            set { _references = value; }
        }
        public List<Member> MemberProperties
        {
            get { return _memberProperties; }
        }
        public override void EdgesCalculated()
        {
            foreach(CodeReference r in _references)
            {
                
                if (r.Source.Project == r.Target.Project)
                {
                    r.Source.ProjectDependency = true;
                    r.Target.ProjectDependency = true;
                }
            }
            if (_methods <= 0)
                _areMethodsStatic = false;
            if (_fields <= 0)
                _areFieldsStatic = false;
            if (_staticMethods > 0)
                _containsStatic = true;
            if (_methods > 0 && _staticMethods < _methods)
                _containsNonStatic = true;
            if (_containsNonStatic == false && _containsStatic == false)
                _containsNonStatic = true;

            _references = null;
        }
        public bool isFullyInitialized()
        {
            if (_name != null && _namespace != null && _fullName != null)
                return true;
            return false;

        }
        public override void AddReference(ReferenceObject c, ReferenceType type, bool isPublished, bool isStatic,SourceFile source)
        {
            //Tracer.WriteTrace("Reference to: {0}, type: {1}", c.Name, type);
            BaseClass cl = (BaseClass)c;
            CodeReference r = new CodeReference(this, c, type, isPublished,isStatic,source);
            
            _references.Add(r);
        }
        public override void AddReference(CodeReference r)
        {
            BaseClass s = (BaseClass)r.Source;
            if (this != s)
                return;

            BaseClass t = (BaseClass)r.Target;
            //Tracer.WriteTrace("Reference to: {0}, type: {1}", t.Name, r.Type);
            _references.Add(r);
        }
        public void AddMemberProperty(string t, string n,bool s)
        {
            Member temp = new Member(t, n, s);
            if (!_memberProperties.Contains(temp))
                _memberProperties.Add(temp);
        }
        public void AddMemberField(string t, string n, bool s)
        {
            Member temp = new Member(t, n, s);
            if (!_memberFields.Contains(temp))
                _memberFields.Add(temp);
        }
        public void AddMethodReturnType(string method,string returnType)
        {
            if(!_methodReturnTypes.ContainsKey(method))
                _methodReturnTypes.Add(method, returnType);
        }
        public string GetReturnFromMethod(string method)
        {
            if (_methodReturnTypes.ContainsKey(method))
                return _methodReturnTypes[method];
            else
                return null;
        }
        public bool ContainsMember(string member)
        {
            return _memberProperties.Where(m => m.Name == member).Count() > 0;
        }
        public Member GetMember(string name)
        {
            return _memberProperties.Where(m => m.Name == name).First();
        }
        public bool ContainsField(string field)
        {
            return _memberFields.Where(m => m.Name == field).Count() > 0;
        }
        public Member GetField(string field)
        {
            return _memberFields.Where(m => m.Name == field).First();
        }
        public override string ToString()
        {
            return _fullName;
        }
    }
    public abstract class ReferenceObject : ReferenceObjectMethods
    {
        protected int _fanIn, _fanOut, _methods, _staticMethods, _constructors, _dependsOnScc,
                    _fields, _staticFields, _publicStaticFields, _privateFields, _solFields, _properties,
                    _sccCount, _reachability;
        protected bool _areMethodsStatic, _areFieldsStatic, _isInterface, _isEnum, _isAbstract, _isPublished, _containsStatic, _containsNonStatic, _projectDependency;
        protected string _name, _fullName, _project;
        protected ReferenceObject _parent;
        protected DependencyVertex _vertex;
        protected List<CodeReference> _references;

        public ReferenceObject()
        {
            _methods = 0;
            _staticMethods = 0;
            _constructors = 0;
            _fanIn = 0;
            _fanOut = 0;
            _staticFields = 0;
            _publicStaticFields = 0;
            _privateFields = 0;
            _dependsOnScc = 0;
            _properties = 0;
            _areFieldsStatic = true;
            _areMethodsStatic = true;
            _isPublished = true;
            _projectDependency = false;
            _containsStatic = false;
            _containsNonStatic = false;
            _isInterface = false;
            _isAbstract = false;
            _isEnum = false;
            _project = "";
        }
        public ReferenceObject(string n)
            : this()
        {
            _name = n;
        }
        public string Name
        {
            get { return _name; }
            set { _name = value; }
        }
        public string FullName
        {
            get { return _fullName; }
        }
        public string Project
        {
            get
            {
                if (_project != "" || _parent == null) 
                    return _project;
                else 
                    return _parent._project;
            }
            set { _project = value; }
        }
        public int FanOut
        {
            get { return _fanOut; }
        }
        public int FanIn
        {
            get { return _fanIn; }
        }
        public int MethodCount
        {
            get { return _methods; }
        }
        public int StaticMethods
        {
            get { return _staticMethods; }
        }
        public int Constructors
        {
            get { return _constructors; }
        }
        public int MemberFields
        {
            get { return _solFields; }
        }
        public int Fields
        {
            get { return _fields; }
        }
        public int StaticFields
        {
            get { return _staticFields; }
        }
        public int PublicStaticFields
        {
            get { return _publicStaticFields; }
        }
        public int PrivateFields
        {
            get { return _privateFields; }
        }
        public int Properties
        {
            get { return _properties; }
        }
        public int Reachability
        {
            get { return _reachability; }
            set { _reachability = value; }
        }
        public int DependsOnSCC
        {
            get { return _dependsOnScc; }
            set { _dependsOnScc = value; }
        }
        public int SccCount
        {
            get { return _sccCount; }
            set { _sccCount = value; }
        }
        public bool AreAllFieldsStatic
        {
            get { return _areFieldsStatic; }
            set { _areFieldsStatic = value; }
        }
        public bool AreAllMethodsStatic
        {
            get { return _areMethodsStatic; }
            set { _areMethodsStatic = value; }
        }
        public bool ProjectDependency
        {
            get { return _projectDependency; }
            set { _projectDependency = value; }
        }
        public bool IsEnum
        {
            get { return _isEnum; }
            set { _isEnum = value; }
        }
        public bool IsInterface
        {
            get { return _isInterface; }
            set { _isInterface = value; }
        }
        public bool IsAbstract
        {
            get { return _isAbstract; }
            set { _isAbstract = value; }
        }
        public bool IsPublished
        {
            get { return _isPublished; }
            set { _isPublished = value; }
        }
        public bool ContainsStatic
        {
            get { return _containsStatic; }
        }
        public bool ContainsNonStatic
        {
            get { return _containsNonStatic; }
        }

        public void IncrMethods()
        {
            _methods++;
            if (_parent != null) _parent.IncrMethods();
        }
        public void IncrStaticMethods()
        {
            _staticMethods++;
            if (_parent != null) _parent.IncrStaticMethods();
        }
        public void IncrConstructors()
        {
            _constructors++;
            if (_parent != null) _parent.IncrConstructors();
        }
        public void IncrFanOut()
        {
            _fanOut++;
            //if (_parent != null) _parent.IncrFanOut();
        }
        public void IncrFanIn()
        {
            _fanIn++;
            //if (_parent != null) _parent.IncrFanIn();
        }
        public void IncrFields()
        {
            _fields++;
            if (_parent != null) _parent.IncrFields();
        }
        public void IncrSolFields()
        {
            _solFields++;
            if (_parent != null) _parent.IncrSolFields();
        }
        public void IncrStaticFields()
        {
            _staticFields++;
            if (_parent != null) _parent.IncrStaticFields();
        }
        public void IncrPublicStaticFields()
        {
            _publicStaticFields++;
            if (_parent != null) _parent.IncrPublicStaticFields();
        }
        public void IncrPrivateFields()
        {
            _privateFields++;
            if (_parent != null) _parent.IncrPrivateFields();
        }
        public void IncrProperties()
        {
            _properties++;
            if (_parent != null) _parent.IncrProperties();
        }
        public void DecrMethods()
        {
            _methods++;
            if (_parent != null) _parent.DecrMethods();
        }
        public void DecrFanOut()
        {
            _fanOut--;
            //if (_parent != null) _parent.DecrFanOut();
        }
        public void DecrFanIn()
        {
            _fanIn--;
            //if(_parent != null)_parent.DecrFanIn();
        }
        public virtual void AddReference(ReferenceObject c, ReferenceType type, bool isPublished, bool isStatic, SourceFile source) { }
        public virtual void AddReference(CodeReference r) { }
        public virtual void EdgesCalculated(){}
    }
    public interface ReferenceObjectMethods
    {
        void AddReference(ReferenceObject c, ReferenceType type, bool isPublished, bool isStatic, SourceFile source);
        void AddReference(CodeReference r);
        void EdgesCalculated();
    }
    public class SourceFile : ReferenceObject, IEquatable<SourceFile>
    {
        private string _dir;

        private List<string> _usingNamespaces;

        public SourceFile(string n)
            : base(n)
        {
            _usingNamespaces = new List<string>();
            _references = new List<CodeReference>();
            if (n.Contains('\\'))
            {
                string[] arr = n.Split('\\');

                _dir = n.Substring(0, n.LastIndexOf('\\'));
                _name = arr[arr.Count() - 1];
            }
            else
            {
                _dir = "";
                _name = n;
            }
            
            _fullName = n;
            _vertex = new DependencyVertex(this);
        }
        
        public List<string> UsingNamespaces
        {
            get { return _usingNamespaces; }
            set { _usingNamespaces = value; }
        }
        public string Namespace
        {
            get { return _dir; }
        }
        public List<CodeReference> References
        {
            get { return _references; }
            set { _references = value; }
        }
        public DependencyVertex Vertex
        {
            get { return _vertex; }
        }
        public bool ContainsUsing(string ns)
        {
            return _usingNamespaces.Contains(ns);
        }
        public void AddUsingNamespace(string ns)
        {
            _usingNamespaces.Add(ns);
        }
        public override void AddReference(ReferenceObject c, ReferenceType type, bool isPublished, bool isStatic,SourceFile s)
        {
            //Tracer.WriteTrace("Reference to: {0}, type: {1}", c.Name, type);
            CodeReference r = new CodeReference(this, c, type, isPublished, isStatic);
            if (c.Project == _project)
                _projectDependency = false;
            _references.Add(r);
        }
        public override void EdgesCalculated()
        {
            if (_methods <= 0)
            {
                _areMethodsStatic = false;
                _areFieldsStatic = false;
            }
        }

        public bool Equals(SourceFile other)
        {
            if (this._fullName == other._fullName)
                return true;
            else
                return false;
        }
    }
    public class Project
    {
        private int _abstractClasses, _efferentCouplings, _afferentCouplings, _couplings, _referenceCount, _classTypeCount;
        private double _relationalCohesion, _instability, _abstractness, _distFromMain;


        public Project()
        {
        }
        public Project(int abstractClasses,int efferentCouplings,int afferentCouplings,int referenceCount,int classTypeCount,int couplings)
        {
            _abstractClasses = abstractClasses;
            _efferentCouplings = efferentCouplings;
            _afferentCouplings = afferentCouplings;
            _referenceCount = referenceCount;
            _classTypeCount = classTypeCount;
            _couplings = couplings;

            _relationalCohesion = (_referenceCount + 1) / _classTypeCount;
            _instability = _efferentCouplings / _couplings;
            _abstractness = _abstractClasses / _classTypeCount;
            _distFromMain = _abstractness + _instability;
        }


        public int AbstractClasses
        {
            get { return _abstractClasses; }
            set { _abstractClasses = value; }
        }
        public int EfferentCouplings
        {
            get { return _efferentCouplings; }
            set { _efferentCouplings = value; }
        }
        public int AfferentCouplings
        {
            get { return _afferentCouplings; }
            set { _afferentCouplings = value; }
        }

        public int Couplings
        {
            get { return _couplings; }
            set { _couplings = value; }
        }

        public int ReferenceCount
        {
            get { return _referenceCount; }
            set { _referenceCount = value; }
        }

        public int ClassTypeCount
        {
            get { return _classTypeCount; }
            set { _classTypeCount = value; }
        }
        public double DistanceFromMainSequence
        {
            get { return _distFromMain; }
            set { _distFromMain = value; }
        }

        public double Abstractness
        {
            get { return _abstractness; }
            set { _abstractness = value; }
        }

        public double Instability
        {
            get { return _instability; }
            set { _instability = value; }
        }

        public double RelationalCohesion
        {
            get { return _relationalCohesion; }
            set { _relationalCohesion = value; }
        }
    }
}
