﻿#region License
/*
Copyright 2014 Christian Thurmann-Nielsen

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using GraphController.Utility;
using GraphController.Logic;
using GraphController.Common;
using GraphController.GUI;
using java.util;
using System.Globalization;

namespace GraphController.Controller
{
    class LogicController
    {
        private GraphController.GUI.RefactoredEdges _reportEdges;
        private GraphLogic _currentLogic;
        private SimulationResult _currentResult;
        private string _report, _lastSimulation, _programDir;
        private bool _simulationFinished;

        public LogicController(MapModel m,DependencyGraph g)
        {
            _currentLogic = new GraphLogic(m, g);
        }

        public void SetInputParameters(int topKFanin, int topKCrss, int topKScc, string mainClass)
        {
            _currentLogic.TopKFanin = topKFanin;
            _currentLogic.TopKCrss = topKCrss;
            _currentLogic.TopKScc = topKScc;
            if (mainClass == "" || mainClass == null)
                throw new ArgumentNullException();
            else
                _currentLogic.MainClass = mainClass;
        }
        public void SetTable1SortingPriority(string s)
        {
            _currentLogic.SetSortingPriority(s);
        }
        public int GetTopKFanin()
        {
            return _currentLogic.TopKFanin;
        }
        public int GetTopKCRSS()
        {
            return _currentLogic.TopKCrss;
        }
        public int GetTopKSCC()
        {
            return _currentLogic.TopKScc;
        }
        public string GetMainClass()
        {
            return _currentLogic.MainClass;
        }
        public SimulationResult GetCurrentResult()
        {
            return _currentResult;
        }
        public void SetNumOfRefactoring(int number)
        {
            _currentLogic.NumOfRefactoring = number;
        }
        public void SetReductionRate(double rate)
        {
            _currentLogic.ReductionRate = (int)rate;
        }
        public List<CycleData> GetCycleData()
        {
            return _currentLogic.CycleData;
        }
        public List<SingleComponentRefactoringData> CalculateSingleComponentCandidateData()
        {
            List<SingleComponentRefactoringData> data = new List<SingleComponentRefactoringData>();
            List s = _currentLogic.GetSingleComponentCandidateData(); // "source,target,crss,fanin,fanout,scc,reductionrate,decision"
            object[] objects = s.toArray();
            _lastSimulation = "Single Component Refactoring: \nComponents: " + s.size();

            foreach (object o in objects)
            {
                SingleComponentRefactoringData single;
                string line = (string)o;
                Tracer.WriteTrace("\tString from tool: {0}", line);
                string[] arr = line.Split(',');
                try
                {
                    string source = arr[0];
                    string target = arr[1];
                    int reach = Convert.ToInt32(arr[2]);
                    int fanin = Convert.ToInt32(arr[3]);
                    int fanout = Convert.ToInt32(arr[4]);
                    int scc = Convert.ToInt32(arr[5]);
                    double reduction = double.Parse(arr[6], CultureInfo.InvariantCulture);
                    string decision = arr[7];
                    single = new SingleComponentRefactoringData(source, target, fanin, fanout, reach, scc, reduction, decision);
                    data.Add(single);
                }
                catch (ArgumentNullException)
                {
                    continue;
                }
                catch (FormatException e)
                {
                    Tracer.WriteTrace("FormatException: {0}", e.StackTrace);
                    continue;
                }
                catch (OverflowException)
                {
                    continue;
                }

            }
            return data;
        }
        public List<SystemRefactoringData> CalculateSystemCandidateData(double rate)
        {
            List<SystemRefactoringData> data = new List<SystemRefactoringData>();
            List s = _currentLogic.GetSystemCandidateData(rate); //"class,crss,fanin,scc,reducedProp,Refactor_decision"
            object[] objects = s.toArray();
            _lastSimulation = "System Refactoring: \nComponents: " + s.size();

            foreach (object o in objects)
            {
                SystemRefactoringData single;
                string line = (string)o;
                Tracer.WriteTrace("\tString from tool: {0}", line);
                string[] arr = line.Split(',');
                try
                {
                    string source = arr[0];
                    int reach = Convert.ToInt32(arr[1]);
                    int fanin = Convert.ToInt32(arr[2]);
                    int scc = Convert.ToInt32(arr[3]);
                    double reduction = double.Parse(arr[4], CultureInfo.InvariantCulture);
                    string decision = arr[5];
                    single = new SystemRefactoringData(source, reach, fanin, scc, reduction, decision);
                    data.Add(single);
                }
                catch (ArgumentNullException)
                {
                    continue;
                }
                catch (FormatException e)
                {
                    Tracer.WriteTrace("FormatException: {0}", e.StackTrace);
                    continue;
                }
                catch (OverflowException)
                {
                    continue;
                }

            }
            return data;
        }
        public SimulationResult SimulateCycleRefactoring(List<CycleRefactoringData> data)
        {
            _simulationFinished = true;

            _currentResult = _currentLogic.SimulateCycleRefactoring(data);
            return _currentResult;// = _logic.SimulateCycleRefactoring(data);
        }
        public SimulationResult SimulateSingleComponentRefactoring(List<SingleComponentRefactoringData> data)
        {
            _simulationFinished = true;
            _currentResult = _currentLogic.SimulateSingleComponentRefactoring(data);
            return _currentResult;
        }
        public SimulationResult SimulateSystemRefactoring(int number)
        {
            _simulationFinished = true;
            _currentResult = _currentLogic.SimulateSystemRefactoring(number);
            return _currentResult;
        }
        private void GenerateSubGraph()
        {
            List l = _currentResult.RefactoredEdges;
            Tracer.WriteTrace("Refactored Edges: {0}", l.size());
            StringBuilder sb = new StringBuilder();
            sb.Append("Report: \n");
            sb.Append(_lastSimulation + "\n\n");
            for (int i = 0; i < l.size(); i++)
            {
                string o = (string)l.get(i).ToString();
                sb.Append(o + "\n");
                Tracer.WriteTrace("\tindex {0}: {1}", i, o);
            }
            _report = sb.ToString();
        }
        public GraphController.GUI.RefactoredEdges GetReport()
        {
            if (_reportEdges.Size == 0)
                return null;

            //GenerateSubGraph();
            return _reportEdges;
        }
        public List<CycleRefactoringData> CalculateCycleCandidateData(CycleData d,bool flag)
        {
            List<CycleRefactoringData> data = new List<CycleRefactoringData>();
            d.PrindCycleSet();
            List s = _currentLogic.GetCycleCandidates(d.CycleSet, flag); //"from,to,crss,scc,reducedRate,Refactor_decision" 
            object[] objects = s.toArray();
            _lastSimulation = "Cycle Refactoring: \nSize: " + d.Size + "\tMain Class: " + d.MainClass.FullName;

            Tracer.WriteTrace("Calculate Cycle: {0}", objects.Count());
            foreach (object o in objects)
            {
                CycleRefactoringData cycle;
                string line = (string)o;
                Tracer.WriteTrace("Calculate Cycle: line: {0}", line);
                string[] arr = line.Split(',');
                try
                {
                    string source = arr[0];
                    string target = arr[1];
                    int reach = Convert.ToInt32(arr[2]);
                    int scc = Convert.ToInt32(arr[3]);
                    double reduction = double.Parse(arr[4], CultureInfo.InvariantCulture);
                    string decision = arr[5];
                    cycle = new CycleRefactoringData(source, target, reach, scc, reduction, decision);
                    data.Add(cycle);
                }
                catch (ArgumentNullException)
                {
                    continue;
                }
                catch (FormatException)
                {
                    continue;
                }
                catch (OverflowException)
                {
                    continue;
                }

            }
            Tracer.WriteTrace("Returning with {0} cycleobjects", data.Count);
            return data;
        }
    }
}
