﻿#region License
/*
Copyright 2014 Christian Thurmann-Nielsen

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Text;
using System.Threading.Tasks;
using GraphController;
using GraphController.Common;
using GraphController.Utility;

namespace Parser
{
    public class DependencyCalculator
    {
        private Dictionary<String,Namespace> _namespaceList;
        private Dictionary<String, BaseClass> _classList;
        private Dictionary<String, SourceFile> _sourceFileList;
        private string _currentSourceFile;
        private Class _currentClass;
        private static String GLOBAL = "GLOBALNAMESPACE";
        private const int LOADING = (int)(Int32.MaxValue * 0.4);
        private MainController _model;

        public DependencyCalculator(MainController model)
        {
            _model = model;
            _namespaceList = new Dictionary<string, Namespace>();
            _classList = new Dictionary<String, BaseClass>();
            _sourceFileList = new Dictionary<String, SourceFile>();
            Namespace global = new Namespace(GLOBAL);
            _namespaceList.Add(GLOBAL, global);
        }
        public string CurrentSourceFile
        {
            get{ return _currentSourceFile;}
            set {   _currentSourceFile = value; 
                    if(!_sourceFileList.ContainsKey(value)) 
                        _sourceFileList.Add(value, new SourceFile(value));
                }
        }
        
        public void AddNamespaceToGraph(string ns)
        {
            if (!_namespaceList.ContainsKey(ns))
            {
                _namespaceList.Add(ns, new Namespace(ns));
            }
        }
        public Namespace GetGlobalNamespace()
        {
            return _namespaceList[GLOBAL];
        }
        public Namespace GetNamespace(string ns)
        {
            return _namespaceList[ns];
        }
        public SourceFile GetCurrentSourceFile()
        {
            return _sourceFileList[_currentSourceFile];
        }
        #region Private methods

        
        #endregion
        private void PrintNamespaces()
        {
            Tracer.WriteTrace("PrintNamespaces: ");
            foreach (Namespace n in _namespaceList.Values)
            {
                Tracer.WriteTrace("    {0}",n.Name);
                foreach (Class c in n.Classes)
                {
                    Tracer.WriteTrace("         {0}", c.Name);
                }
            }
        }
        private void PrintClasses()
        {
            Tracer.WriteTrace("PrintClasses: ");
            foreach (Class c in _classList.Values)
            {
                Tracer.WriteTrace("    {0}, size: {1}", c.FullName,c.References.Count);
            }
        }
        private string[] SplitNamespace(string s)
        {
            string[] arr = new string[2];
            if (s.Contains('.'))
            {
                int index = s.LastIndexOf('.');
                arr[0] = s.Substring(0, index);
                arr[1] = s.Substring(index + 1);
                return arr;
            }
            else
                return null;
        }
        private CodeReference AddEdge(BaseClass c, string fullName, CodeReference r)
        {
            //Tracer.WriteTrace("Adding Edge: found class: {0} , {1}, {2}", c.FullName, fullName,r.Type);
            BaseClass target = null;
            if( (target = _classList[fullName]) == null)
                return null;

            CodeReference reference;
            if (r.Type == ReferenceType.BASE && target.IsInterface)
            {
                reference = new CodeReference(c, target, ReferenceType.INTERFACE);
                _model.AddEdgeToModel(c, target, reference);
                //original.AddEdgeToGraph(c.Vertex, target.Vertex, ReferenceType.INTERFACE);
                //copy.AddEdgeToGraph(c.Vertex, target.Vertex, ReferenceType.INTERFACE);
                return reference;
            }

            reference = new CodeReference(c, target, r.Type);
            _model.AddEdgeToModel(c, target, r);
            //original.AddEdgeToGraph(c.Vertex, target.Vertex, r.Type);
            //_namespaceList[c.Namespace.Name].Graph.AddEdgeToGraph(new MyEdgeBase(c.Vertex, reference.Vertex));
            return reference;
        }

        /// <summary>
        /// This algorithm loops through every CodeReference found in a class and finds its corresponding
        /// Class in a Dictionary of classes. Every CodeReference includes a Class where only the name is 
        /// is set. This name is gathered from the parser and is set to the identifier name of a Node in the
        /// abstract syntax tree. The full name of the class is then calculated based on the using statements
        /// and the current namespace the identifier node is found in.
        /// </summary>
        public void CalculateEdges()
        {
            int inc = 0;
            _classList = _model.ClassList;
            if (_classList.Count <= 0)
                return;

            inc = (LOADING / _classList.Count);

            foreach (BaseClass c in _classList.Values)
            {
                List<CodeReference> newReferences = new List<CodeReference>();
                foreach (CodeReference r in c.References)
                {
                    FindDependency(c,newReferences,r);
                }
                c.References = newReferences;
                _model.FinaliseClass(c);
                _model.ProgressBar.IncrementProgress(inc);
            }
        }
        private void FindDependency(BaseClass c,List<CodeReference> newReferences ,CodeReference r)
        {
            Class c2 = (Class)r.Target;

            if (c2.Namespace == null)
            {
                string namesp = "";
                string name = "";
                string[] arr = SplitNamespace(c2.Name);
                string[] namespaces = null;

                if (arr == null) //c2.Name only contains the Class name, 
                //so the name is either from using statements or within 
                //the namespace of c, or in the global namespace
                {
                    name = c2.Name;
                    if (c.Namespace.IsClassInNamespace(name))
                    {
                        string fullName = c.Namespace.Name + "." + c2.Name;
                        if (c.FullName == fullName)
                            return;
                        newReferences.Add(AddEdge(c, fullName, r));
                        return;
                    }
                    else
                    {
                        namespaces = c.FullName.Split('.');
                        if (namespaces != null)
                        {
                            string acc = "";
                            foreach (string s in namespaces)
                            {
                                if (acc == "")
                                    acc = s;
                                else
                                    acc += '.' + s;

                                CodeReference temp;
                                if ((temp = CheckNamespace(acc, c, c2.Name, r)) != null)
                                {
                                    newReferences.Add(temp);
                                    return;
                                }
                            }
                        }
                        if (c.GetType() == typeof(PartialClass))
                        {
                            CheckNamespaces(c, c2.Name, r, newReferences, r.SourceFile.UsingNamespaces);
                        }
                        else
                        {
                            Class cl = (Class)c;
                            CheckNamespaces(c, c2.Name, r, newReferences, cl.SourceFile.UsingNamespaces);
                        }

                    }
            
                    if (_namespaceList[GLOBAL].IsClassInNamespace(name))
                    {
                        string fullName = _namespaceList[GLOBAL].Name + "." + c2.Name;
                        if (c.FullName == fullName)
                            return;
                        newReferences.Add(AddEdge(c, fullName, r));
                        return;
                    }
                }
                else
                {
                    name = arr[1];
                    namesp = arr[0];

                    namespaces = namesp.Split('.');
                    string temp = c.Namespace.Name + "." + namesp;

                    CodeReference cr;
                    if ((cr = CheckNamespace(temp, c, name, r)) != null)
                    {
                        newReferences.Add(cr);
                        return;
                    }
                    else if (namespaces != null)
                    {
                        string acc = "";
                        foreach (string s in namespaces)
                        {
                            if (acc == "")
                                acc = s;
                            else
                                acc += "." + s;

                            CodeReference cr2;
                            if ((cr2 = CheckNamespace(acc, c, name, r)) != null)
                            {
                                newReferences.Add(cr2);
                                return;
                            }
                        }
                    }
                }

            }
        }
        private void CheckNamespaces(BaseClass c,string refClassName,CodeReference r,List<CodeReference> l, List<string> namespaces)
        {
            foreach (string ns in namespaces)
            {
                CodeReference temp;
                if ((temp = CheckNamespace(ns, c, refClassName, r)) != null)
                    l.Add(temp);
            }
        }
        private CodeReference CheckNamespace(string ns,BaseClass c, string refClassName,CodeReference r)
        {
            if (_namespaceList.ContainsKey(ns) && _namespaceList[ns].IsClassInNamespace(refClassName))
            {
                string fullName = _namespaceList[ns].Name + "." +refClassName;
                if (c.FullName == fullName)
                    return null;
                return AddEdge(c, fullName, r);
            }
            return null;
        }
        //private void CalculateNamespaceInSourceFile(ref label,string ns,BaseClass c, BaseClass c2,List<CodeReference> l)
        //{
        //    if (_namespaceList.ContainsKey(ns) && _namespaceList[ns].IsClassInNamespace(c2.Name))
        //    {
        //        string fullName = _namespaceList[ns].Name + "." + c2.Name;
        //        if (c.FullName == fullName)
        //            goto end_of_ref;
        //        l.Add(AddEdge(c, fullName, r));
        //        goto end_of_ref;
        //    }
        //}
        public string[] GetNamespaces()
        {
            return _namespaceList.Keys.ToArray();
        }
        public DependencyGraph GetNamespaceGraph(string ns)
        {
            return _namespaceList[ns].Graph;
        }
        internal void AddUsingNamespaceToSourcefile(List<string> l)
        {
            _sourceFileList[_currentSourceFile].UsingNamespaces = l;
        }
        public void AddUsingNamespaceToSourcefile(string s)
        {
            _sourceFileList[_currentSourceFile].UsingNamespaces.Add(s);
        }
        /*internal void AddUsingNamespaceToGraph(string cf,string ns,List<string> l)
        {
            _namespaceList[ns].AddUsingNamespace(cf, l);
        }*/
       
    }
    
}
