﻿#region License
/*
Copyright 2014 Christian Thurmann-Nielsen

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/
#endregion

using System;
using System.Reflection;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.IO;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Windows.Threading;
using System.Text;
using System.Threading.Tasks;
using ICSharpCode.NRefactory.CSharp.Analysis;
using ICSharpCode.NRefactory.CSharp;
using System.Collections;
using GraphController;
using GraphController.Common;
using GraphController.Utility;

namespace Parser
{
    
    public class DependencyFinder
    {
        private SyntaxTree tree;
        private string _currentNamespace, project;
        private Stack<String> previousNamespace, previousClass;
        private BaseClass _currentClass;
        private Dictionary<string, string> _currentIdentifiers, _methodIdentifiers, _overridenIdentifiers;
        private DependencyCalculator data;
        private int treeLevel, loadingIncrement;
        private string sourceFile;
        private List<string> usingNamespacesInSourceFile;
        private ReferenceType _currentType;
        private const int LOADING = (int)(Int32.MaxValue * 0.6);
        private bool debug;
        private bool _isPublished, _isMethodStatic, _isMethod;
        private MainController _model;
        private BackgroundWorker _worker;
        private EventHandler guiDelegate;
        private List<ClassFile> _classFiles;
        private Dispatcher _guiDisp;

        public DependencyFinder(MainController model, Dispatcher d)
        {
            
            Trace.WriteLine(string.Format(CultureInfo.CurrentCulture, "DepFinder, constructor"));
            _model = model;
            _model.ProgressBar.Update();
            treeLevel = 0;
            data = new DependencyCalculator(_model);
            _isPublished = false;
            _isMethodStatic = false;
            _isMethod = false;
            previousNamespace = new Stack<string>();
            previousClass = new Stack<string>();
            _currentNamespace = data.GetGlobalNamespace().Name;
            _worker = new BackgroundWorker();
            _worker.WorkerReportsProgress = true;
            _worker.WorkerSupportsCancellation = true;
            _worker.ProgressChanged += new ProgressChangedEventHandler(_worker_ProgressChanged);
            _worker.DoWork += new DoWorkEventHandler(_worker_DoWork);
            _worker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(_worker_RunWorkerCompleted);
            _guiDisp = d;
        }

        public void parseClassFiles(List<ClassFile> list)
        {
            Trace.WriteLine(string.Format(CultureInfo.CurrentCulture, "DepFinder, parseClassFiles()"));
            //queue.Enqueue(Task.Factory.StartNew(() => ClassFileFinder.FindAllScriptFiles(dirPath, sourceFiles, queue)));

            //while (queue.Count > 0)
            //{
            //    Trace.WriteLine(string.Format(CultureInfo.CurrentCulture, "DepFinder, parseClassFiles(), waiting for queue: {0}",queue.Count));
            //}

            if (list == null)
            {
                throw new FileLoadException();
            }
            _model.ProgressBar.Reset();
            _model.ProgressBar.Show();
            debug = false;

            loadingIncrement = 0;
            if(list.Count > 0)
                loadingIncrement = (LOADING / list.Count);

            _classFiles = list;

            Parse();

            data.CalculateEdges();
        }
        private void Parse()
        {

            CSharpParser p = new CSharpParser();
            foreach (ClassFile c in _classFiles)
            {
                try
                {
                    project = c.Project;
                    usingNamespacesInSourceFile = new List<string>();
                    TextReader r;
                    try
                    {
                        r = File.OpenText(c.Path);
                    }
                    catch (ArgumentNullException)
                    { continue; }

                    string text = r.ReadToEnd();
                    tree = p.Parse(text, c.Path);
                    if (tree.Children.Count() <= 0)
                        continue;

                    double treeIncrement = loadingIncrement / tree.Children.Count();

                    sourceFile = tree.FileName;
                    data.CurrentSourceFile = sourceFile;
                    data.GetCurrentSourceFile().Project = c.Project;

                    Action action = () => _model.AddSourceFile(data.GetCurrentSourceFile());
                    _guiDisp.Invoke(action); // Or use BeginInvoke
                    //_model.AddSourceFile(data.GetCurrentSourceFile());

                    //Trace.WriteLine(string.Format(CultureInfo.CurrentCulture, "DepFinder, parseClassFiles(), class: {0}", sourceFile));
                    /*var watch = Stopwatch.StartNew();
                    Tracer.WriteTrace("Visitor, tree: {0}", tree.Children.Count());
                    tree.AcceptVisitor(visitor);
                    watch.Stop();
                    elapvisitor += watch.ElapsedMilliseconds;

                    watch = Stopwatch.StartNew();
                    Tracer.WriteTrace("Parse, tree: {0}", tree.Children.Count());*/

                    //watch.Stop();
                    //elapparse += watch.ElapsedMilliseconds;
                    if (debug)
                    {
                        treeLevel++;
                        foreach (var child in tree.Children)
                        {
                            MakeTreeNode(child);
                        }
                    }
                    else
                    {
                        ParseSourceFile(tree);
                        _worker.ReportProgress(loadingIncrement / int.MaxValue);
                    }
                }
                catch (Exception e)
                {
                    e.TargetSite.ToString();
                    continue;
                }

            }
        }
        private void _worker_DoWork(object sender, DoWorkEventArgs e)
        {
            Parse();
        }
        private void _worker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            _model.ProgressBar.IncrementProgress(loadingIncrement);
        }
        private void _worker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            //if ((e.Cancelled == true))
            //{
            //    this.tbProgress.Text = "Canceled!";
            //}

            //else if (!(e.Error == null))
            //{
            //    this.tbProgress.Text = ("Error: " + e.Error.Message);
            //}

            //else
            //{
            //    this.tbProgress.Text = "Done!";
            //}
            data.CalculateEdges();
        }

        string GetNodeTitle(AstNode node)
        {
            StringBuilder b = new StringBuilder();
            if(debug)Tracer.WriteTrace("NewNode({0}) -> {1}: {2}",treeLevel, node.Role.ToString(),node.GetType().Name);
            b.Append(node.Role.ToString());
            b.Append(": ");
            b.Append(node.GetType().Name);
            bool hasProperties = false;
            foreach (PropertyInfo p in node.GetType().GetProperties(BindingFlags.Public | BindingFlags.Instance))
            {
                if (p.Name == "NodeType" || p.Name == "IsNull" || p.Name == "IsFrozen" || p.Name == "HasChildren")
                    continue;
                if (p.PropertyType == typeof(string) || p.PropertyType.IsEnum || p.PropertyType == typeof(bool))
                {
                    if (!hasProperties)
                    {
                        hasProperties = true;
                        b.Append(" (");
                    }
                    else
                    {
                        b.Append(", ");
                    }
                    if (debug) Tracer.WriteTrace("     Property:{0}", p.Name);
                    b.Append(p.Name);
                    b.Append(" = ");
                    try
                    {
                        object val = p.GetValue(node, null);
                        b.Append(val != null ? val.ToString() : "**null**");
                        if (val != null) 
                            if (debug) Tracer.WriteTrace("         value: {0}", val.ToString());
                    }
                    catch (TargetInvocationException ex)
                    {
                        b.Append("**" + ex.InnerException.GetType().Name + "**");
                    }
                }
            }
            if (hasProperties)
                b.Append(")");
            return b.ToString();
        }

        void printRole(AstNode n)
        {
            Tracer.WriteTrace("{0}:{1} -> {2}",n.ToString(), n.Role.ToString(), n.GetType().Name);
        }

        void MakeTreeNode(AstNode node)
        {
            //printRole(node);
            GetNodeTitle(node);
            treeLevel++;
            foreach (AstNode child in node.Children)
            {
                MakeTreeNode(child);
            }
            treeLevel--;
        }
        private void AddUsingDeclaration(AstNode n)
        {
            string ns = CheckProperty(n, "Namespace");
            if (ns != null)
            {
                usingNamespacesInSourceFile.Add(ns);
            }

        }
        private void ParseSourceFile(AstNode n)
        {
            foreach (AstNode child in n.Children)
            {
                if( child.GetType() == typeof(UsingDeclaration)) // find using Namespaces
                {
                    AddUsingDeclaration(child);
                }
                else if (child.GetType() == typeof(NamespaceDeclaration)) // find Namespaces
                {
                    AddNamespaceDeclaration(child);
                }
                else if (child.GetType() == typeof(TypeDeclaration)) // find Namespaces
                {
                    AddTypeDeclaration(child);
                }
            }
        }

        private void ParseNamespace(AstNode n,string ns)
        {
            foreach (AstNode child in n.Children)
            {
                if (child.GetType() == typeof(NamespaceDeclaration)) // find Namespaces
                {
                    previousNamespace.Push(_currentNamespace);
                    AddNamespaceDeclaration(child);
                }
                else if (child.GetType() == typeof(TypeDeclaration))
                {
                    AddTypeDeclaration(child);
                }
                else if (child.GetType() == typeof(CSharpTokenNode) && child.Role.ToString().Equals("}"))
                {
                    if (previousNamespace.Count > 0)
                        _currentNamespace = previousNamespace.Pop();
                    else
                        _currentNamespace = data.GetGlobalNamespace().Name;
                }
            }
            
        }
        private void AddNamespaceDeclaration(AstNode n)
        {
            string ns = CheckProperty(n, "FullName");
            if (ns != null)
            {
                //Tracer.WriteTrace("Found Namespace: {0}", ns);
                _currentNamespace = ns;
                data.AddNamespaceToGraph(ns);
                data.AddUsingNamespaceToSourcefile(usingNamespacesInSourceFile);
                ParseNamespace(n, ns);
            }
        }
        //private void PrintIdentifiers()
        //{
        //    try
        //    {
        //        if(_currentIdentifiers.Count > 0)
        //        {
        //            Tracer.WriteTrace("--Current identifiers in class: {0}", _currentClass.FullName);

        //            foreach (KeyValuePair<string, string> p in _currentIdentifiers)
        //            {
        //                Tracer.WriteTrace("\t--Pair: {0} : {1}", p.Key, p.Value);
        //            }
        //        }
                
        //        if(_methodIdentifiers.Count > 0)
        //        {
        //            Tracer.WriteTrace("\t--Current methodidentifiers in class: {0}", _currentClass.FullName);
        //            foreach (KeyValuePair<string, string> p in _methodIdentifiers)
        //            {
        //                Tracer.WriteTrace("\t\t--Pair: {0} : {1}", p.Key, p.Value);
        //            }
        //        }
                
        //    }
        //    catch(NullReferenceException)
        //    {

        //    }
        //}
        private void AddTypeDeclaration(AstNode n)
        {
            if(_currentClass != null)
                previousClass.Push(_currentClass.FullName);

            TypeDeclaration d = (TypeDeclaration)n;
            BaseClass c;
            Namespace ns = null;
            string fullname = "";
            bool partial = false;
            _isMethod = false;

            if (data.GetNamespace(_currentNamespace) != null)
            {
                ns = data.GetNamespace(_currentNamespace);
            }
            else
            {
                ns = data.GetGlobalNamespace();
            }

            fullname = ns.Name + "." + d.Name;
            //Tracer.WriteTrace("---Class: {0}", fullname);

            if (d.Modifiers.HasFlag(Modifiers.Partial))
                partial = true;
            
            if (_model.ContainsClass(fullname))
            {
                c = _model.GetClass(fullname);
                if(c.GetType() == typeof(PartialClass))
                {
                    //Tracer.WriteTrace("---PartialClass exists: {0}", fullname);
                    PartialClass pc = (PartialClass)c;
                    pc.AddSourcefile(data.GetCurrentSourceFile());
                }
                //else
                    //Tracer.WriteTrace("---Class exists and not Partial: {0}", fullname);
            }
            else
            {
                if (partial)
                {
                    Tracer.WriteTrace("---Creating partial class: {0}", fullname);
                    c = new PartialClass(d.Name, ns, data.GetCurrentSourceFile(), project);
                }
                else
                {
                    c = new Class(d.Name, ns, data.GetCurrentSourceFile(), project);
                }

            }
            
            _currentClass = c;
            _currentIdentifiers = new Dictionary<string, string>();
            _overridenIdentifiers = new Dictionary<string, string>();
            _methodIdentifiers = new Dictionary<string, string>();
            
            if (d.ClassType.HasFlag(ClassType.Interface) && !d.ClassType.HasFlag(ClassType.Enum))
                c.IsInterface = true;
            else if (d.ClassType.HasFlag(ClassType.Enum))
                c.IsEnum = true;

            if (d.Modifiers.HasFlag(Modifiers.Abstract))
                c.IsAbstract = true;


            _model.AddClass(c);
            CheckForBaseClasses(n);
            ParseClass(n);
            
        }
        private void FoundSimpleType(AstNode n)
        {
            string role = n.Role.ToString();
            switch (role)
            {
                case RoleStrings.BASETYPE:
                    AddBaseTypeReference(n);
                    break;
            }
        }
        private void AddBaseTypeReference(AstNode n)
        {
            Class node = new Class();
            if ((node.Name = CheckProperty(n, "Identifier")) != null)
            {
                _currentClass.AddReference(node, _currentType, _isPublished,_isMethodStatic,data.GetCurrentSourceFile());
            }
        }
        private void ParseClass(AstNode n)
        {
            foreach (AstNode child in n.Children)
            {
                if (child.GetType() == typeof(TypeDeclaration))
                {
                    AddTypeDeclaration(child);
                }
                else if(FindMembersInClass(child))
                    continue;
                else if(FindMethodsInClass(child))
                    continue;
                else if(child.GetType() == typeof(CSharpTokenNode) && child.Role.ToString().Equals("}"))
                {
                    _currentClass = null;
                    if(previousClass.Count > 0)
                    {
                        string cl = previousClass.Pop();
                        _currentClass = _model.GetClass(cl);
                    }

                }
            }
        }
        private void CheckForBaseClasses(AstNode n)
        {
            _currentType = ReferenceType.BASE;
            foreach (AstNode child in n.Children)
            {
                if (child.Role.ToString().Equals("BaseType") && (child.GetType() == typeof(SimpleType) || child.GetType() == typeof(MemberType)))
                {
                    Class node = new Class();
                    if (child.GetType() == typeof(MemberType))
                    {
                        if ((node.Name = FindClassName(child)) != null)
                        {
                            node.Name += "." + CheckProperty(child,"MemberName") ;
                            _currentClass.AddReference(node, _currentType, _isPublished, _isMethodStatic, data.GetCurrentSourceFile());
                        }

                    } 
                    else if ((node.Name = CheckProperty(child,"Identifier")) != null)
                    {
                        _currentClass.AddReference(node, _currentType, _isPublished, _isMethodStatic, data.GetCurrentSourceFile());
                    }

                }
            }
        }

        private bool FindMembersInClass(AstNode n)
        {
            if (n.Role.Equals(Roles.TypeMemberRole) && ( n.GetType() == typeof (FieldDeclaration) || n.GetType() == typeof(PropertyDeclaration)))
            {
                _isMethod = false;
                bool isStatic = false;
                if(n.GetType() == typeof(FieldDeclaration))
                {
                    _currentType = ReferenceType.MEMBERFIELD;
                    FieldDeclaration f = (FieldDeclaration)n;
                    isStatic = f.Modifiers.HasFlag(Modifiers.Static);
                    
                    FindModifiersInField(f); 
                    _currentClass.IncrFields();
                    
                    foreach (VariableInitializer child in f.Variables)
                    {
                        Class node = new Class();
                        if ((node.Name = FindClassName(n)) != null)
                        {
                            VariableInitializer i = (VariableInitializer)child;
                            CodeReference r = new CodeReference(_currentClass, node, _currentType, false, false, data.GetCurrentSourceFile());
                            _currentClass.AddReference(r);
                            _currentClass.AddMemberField(node.Name, child.Name, isStatic);

                            if (!_currentIdentifiers.ContainsKey(i.Name))
                                _currentIdentifiers.Add(i.Name, node.Name);
                            else
                            {
                                _currentIdentifiers.Remove(i.Name);
                                _currentIdentifiers.Add(i.Name, node.Name);
                            }
                            foreach (AstNode varInitChild in i.Children)
                            {
                                if (varInitChild.GetType() == typeof(ObjectCreateExpression))
                                {
                                    _currentClass.AddReference(node, ReferenceType.CONSTRUCTORINVOCATION, false, false, data.GetCurrentSourceFile());
                                }
                                else if (CheckChild(varInitChild))
                                    continue;
                            }
                        }
                        else
                        {
                            _currentIdentifiers.Add(child.Name, f.ReturnType.ToString());
                            _currentClass.AddMemberField(f.NodeType.ToString(), child.Name, isStatic);
                        }
                        
                    }
                    
                    return true; 
                }
                else if (n.GetType() == typeof(PropertyDeclaration))
                {
                    _currentType = ReferenceType.MEMBERPROPERTY;
                    PropertyDeclaration p = (PropertyDeclaration)n;

                    _isMethod = true;
                    FindReferencedClassesInMethod(p.Getter);
                    FindReferencedClassesInMethod(p.Setter);
                    _isMethod = false;
                    FindModifiersInField(p);
                    _currentClass.IncrProperties();
                    isStatic = p.Modifiers.HasFlag(Modifiers.Static);
                    _isPublished = p.Modifiers.HasFlag(Modifiers.Public);

                    Class prop = new Class();
                    if((prop.Name = FindClassName(p)) != null)
                    {
                        _currentClass.AddMemberProperty(prop.Name, p.Name, isStatic);
                        CodeReference propref = new CodeReference(_currentClass, prop, ReferenceType.MEMBERPROPERTY, _isPublished, isStatic, data.GetCurrentSourceFile());
                        _currentClass.AddReference(propref);
                        if (!_currentIdentifiers.ContainsKey(p.Name))
                            _currentIdentifiers.Add(p.Name, prop.Name);
                        return true;
                    }
                    else
                        _currentClass.AddMemberProperty(p.ReturnType.ToString(), p.Name, isStatic);
                }
            }
            return false;
        }
        private void AddMemberReference(AstNode n)
        {
            Class node = new Class();
            if ((node.Name = FindClassName(n)) != null)
            {
                _currentClass.AddReference(node, _currentType, _isPublished, _isMethodStatic, data.GetCurrentSourceFile());
            }
        }

        private string FindClassName(AstNode n)
        {
            foreach (AstNode memberChild in n.Children)
            {
                if (memberChild.Role.Equals(Roles.Type) || memberChild.Role.Equals(MemberType.TargetRole))
                {
                    string memberClassName = "";
                    if (memberChild.GetType() == typeof(MemberType))
                    {
                        memberClassName = FindClassName(memberChild) + "." + CheckProperty(memberChild,"MemberName") ;
                        return memberClassName;
                    }
                    else if ( memberChild.GetType() == typeof(SimpleType) && (memberClassName += CheckProperty(memberChild, "Identifier")) != null)
                    {
                        var list = memberChild.Children.Where(node => node.Role.Equals(Roles.TypeArgument));
                        if (list.Count() > 0)
                        {
                            AstNode node = list.First();
                            _currentClass.AddReference(new Class(memberClassName), _currentType, false, false, data.GetCurrentSourceFile());
                            if (node.GetType() == typeof(MemberType))
                                return FindClassName(node) + "." + CheckProperty(node, "MemberName");
                            else
                                return CheckProperty(node,"Identifier");
                        }
                        else
                            return memberClassName;
                    }
                    else if(memberChild.GetType() == typeof(ComposedType))
                    {
                        return FindClassName(memberChild);
                    }
                }

            }
            return null;
        }
        private bool AddVariableInMethod()
        {
            return false;
        }
        private void AddVariable()
        {

        }
        private bool FindMethodsInClass(AstNode n)
        {
            //PrintIdentifiers();
            if (n.Role.Equals(Roles.TypeMemberRole) && ( n.GetType() == typeof (MethodDeclaration) || n.GetType() == typeof(ConstructorDeclaration)) )
            {
                _isMethod = true;
                if (_methodIdentifiers != null)
                {
                    foreach (KeyValuePair<string, string> pair in _methodIdentifiers)
                    {
                        _currentIdentifiers.Remove(pair.Key);

                        if (_overridenIdentifiers.ContainsKey(pair.Key))
                            _currentIdentifiers.Add(pair.Key, _overridenIdentifiers[pair.Key]);
                    }
                }
                //PrintIdentifiers();
                
                _methodIdentifiers = new Dictionary<string, string>();


                if ( n.GetType() == typeof (MethodDeclaration))
                {
                    MethodDeclaration m = (MethodDeclaration)n;
                    string accessMods = FindModifiersInMethod(m);
                    
                    string methodName = CheckProperty(n, "Name");
                    string methodSignature = accessMods + methodName + "(";
                    ParameterDeclaration[] arr = m.Parameters.ToArray();
                    
                    for(int i = 0; i < arr.Length; i++)
                    {
                        ParameterDeclaration pd = arr[i];
                        if (accessMods.Contains("public"))
                            _currentType = ReferenceType.METHODPARAMETERPUB;
                        else
                            _currentType = ReferenceType.METHODPARAMETER;

                        string parameterClass = "";
                        if ((parameterClass = AddReference(pd, ReferenceType.METHODPARAMETER)) != null)
                        {
                            AddIdentifier(pd.Name, parameterClass,true);
                            methodSignature += parameterClass;
                        }
                        else
                            methodSignature += pd.Type;

                        if( !((i + 1) == arr.Length) )
                            methodSignature += ",";
                    }
                    methodSignature += ")";
                    _currentType = ReferenceType.METHODRETURNTYPE;
                    if (m.ReturnType.NodeType.HasFlag(NodeType.Member))
                    {
                        MemberType t = (MemberType)m.ReturnType;
                    } 
                    else if (m.ReturnType.GetType().Name.Equals("PrimitiveType"))
                    {
                        PrimitiveType t = (PrimitiveType)m.ReturnType;
                        //Tracer.WriteTrace("\tMethod: {0} returns {1}", methodSignature,t.ToString());
                    }
                    else if (m.ReturnType.GetType() == typeof(SimpleType))
                    {
                        SimpleType t = (SimpleType)m.ReturnType;
                        string returnType = AddReference(m, ReferenceType.METHODRETURNTYPE);
                        _currentClass.AddMethodReturnType(methodSignature, returnType);
                        //Tracer.WriteTrace("\tMethod: {0} returns {1}", methodSignature, returnType);
                    }
                    else if (m.ReturnType.GetType() == typeof(MemberType))
                    {
                        string returnType = AddReference(m, ReferenceType.METHODRETURNTYPE);
                        _currentClass.AddMethodReturnType(methodSignature, returnType);
                    }
                    
                    _currentClass.IncrMethods();
                }
                else if (n.GetType() == typeof(ConstructorDeclaration))
                {
                    ConstructorDeclaration c = (ConstructorDeclaration)n;
                    _currentClass.IncrConstructors();
                    string accessMods =  FindModifiersInMethod(c);
                        
                    if (accessMods.Contains("public"))
                        _currentType = ReferenceType.METHODPARAMETERPUB;
                    else
                        _currentType = ReferenceType.METHODPARAMETER;

                    ParameterDeclaration[] arr = c.Parameters.ToArray();

                    for (int i = 0; i < arr.Length; i++)
                    {
                        ParameterDeclaration pd = arr[i];

                        string parameterClass = "";
                        if ((parameterClass = AddReference(pd, ReferenceType.METHODPARAMETER)) != null)
                        {
                            AddIdentifier(pd.Name, parameterClass,true);
                        }
                    }
                }

                FindReferencedClassesInMethod(n);
                //PrintIdentifiers();
                return true;
            }
            return false;
        }
        private void AddIdentifier(string name, string c,bool isMethod)
        {
            if (!_currentIdentifiers.ContainsKey(name))
            {
                _currentIdentifiers.Add(name, c);
                if (isMethod)
                {
                    if (_methodIdentifiers.ContainsKey(name))
                        _methodIdentifiers.Remove(name);
                    _methodIdentifiers.Add(name, c);
                }
                    
            }
            else if(isMethod)
            {
                string value = _currentIdentifiers[name];
                if(!_overridenIdentifiers.ContainsKey(name))
                    _overridenIdentifiers.Add(name, value);

                _currentIdentifiers.Remove(name);
                _currentIdentifiers.Add(name, c);
                if (_methodIdentifiers.ContainsKey(name))
                    _methodIdentifiers.Remove(name);
                _methodIdentifiers.Add(name, c);
                
            }
            else
            {
                Tracer.WriteTrace("Pompel????");
            }
        }
        private void FindModifiersInField(EntityDeclaration e)
        {
            Modifiers m = e.Modifiers;
            if (m.HasFlag(Modifiers.Static) && m.HasFlag(Modifiers.Public))
            {
                _currentClass.IncrPublicStaticFields();
                _currentType = ReferenceType.STATICFIELD;
                _currentClass.IncrStaticFields();
            }
            else if (m.HasFlag(Modifiers.Static))
            {
                _currentType = ReferenceType.STATICFIELD;
                _currentClass.IncrStaticFields();
            }
            else
                _currentClass.AreAllFieldsStatic = false;

            if (m.HasFlag(Modifiers.Private))
                _currentClass.IncrPrivateFields();


            if (!m.HasFlag(Modifiers.Public))
                _isPublished = false;
        }
        private string AddReference(AstNode n,ReferenceType t)
        {
            string name;
            if ((name = FindClassName(n)) != null)
            {
                Class c = new Class();
                c.Name = name;
                _currentClass.AddReference(c, t, _isPublished, _isMethodStatic, data.GetCurrentSourceFile());
            }
            return name;
        }
        private string FindModifiersInMethod(EntityDeclaration e)
        {
            string s = "";
            Modifiers mods = e.Modifiers;
            if (mods.HasFlag(Modifiers.Public))
            {
                _isPublished = true;
                s += "public ";
            }
            else if (mods.HasFlag(Modifiers.Protected))
            {
                _isPublished = true;
                s += "protected ";
            }
            else
            {
                _isPublished = false;
                _currentClass.IsPublished = false;
            }


            if (!mods.HasFlag(Modifiers.Static))
            {
                _currentClass.AreAllMethodsStatic = false;
                _isMethodStatic = false;
            }
            else
            {
                _isMethodStatic = true;
                _currentClass.IncrStaticMethods();
                s += "static ";
            }
            return s;
        }
        private void FindReferencedClassesInMethod(AstNode n)
        {
            foreach (AstNode memberChild in n.Children)
            {
                FindReferencedClassInBody(memberChild);
            }
        }
        private void FindReferencedClassInBody(AstNode n)
        {
            _currentType = ReferenceType.METHODBODY;
            if (n.Role.ToString().Equals("Body") && n.GetType().Name.Equals("BlockStatement"))
            {
                RecursiveClassFinder(n);
            }
        }

        #region ParsingMethods for the nodes in the abstract syntax tree
        private bool ParseVariableDeclarationStatement(AstNode n)
        {
            if(n.GetType() == typeof(VariableDeclarationStatement))
            {
                VariableDeclarationStatement v = (VariableDeclarationStatement)n;

                _currentType = ReferenceType.MEMBERFIELD;
                string name; 
                    
                if ((name = FindClassName(n)) != null)
                {
                    Class node = new Class();
                    node.Name = name; 
                    foreach (AstNode varDeclChild in n.Children)
                    {
                        if (varDeclChild.GetType() == typeof(VariableInitializer))
                        {
                            foreach (AstNode varInitChild in varDeclChild.Children)
                            {
                                if (ParseExpressionStatement(varInitChild))
                                    continue;
                                else if (ParseMemberReferenceExpression(varInitChild))
                                    continue;
                                else if (ParseObjectCreateExpression(varInitChild))
                                    continue;
                            }
                            VariableInitializer i = (VariableInitializer)varDeclChild;

                            AddIdentifier(i.Name, node.Name, _isMethod);


                        }
                        else if (CheckChild(varDeclChild))
                            continue;
                        else
                            RecursiveClassFinder(varDeclChild);
                    }
                    _currentClass.AddReference(node, ReferenceType.MEMBERFIELD, _isPublished, _isMethodStatic, data.GetCurrentSourceFile());
                    //Tracer.WriteTrace("\t\tVar: {0}", node.Name);
                    return true;
                }
                else
                {
                    foreach(VariableInitializer i in v.Variables)
                    {
                        AddIdentifier(i.Name.ToString(), v.Type.ToString(),_isMethod);
                    }
                }
            }
            return false;
        }
        private bool ParseObjectCreateExpression(AstNode n)
        {
            bool flag = false;
            if (n.GetType() == typeof(ObjectCreateExpression))
            {
                _currentType = ReferenceType.CONSTRUCTORINVOCATION;
                if(AddReference(n, ReferenceType.CONSTRUCTORINVOCATION) != null)
                    flag = true;
                RecursiveClassFinder(n);
            }
            return flag;
        }
        private bool ParseInvocationExpression(AstNode n)
        {
            if (n.GetType() == typeof(InvocationExpression))
            {
                bool flag = false;
                InvocationExpression i = (InvocationExpression)n;
                foreach (AstNode invChild in n.Children)
                {

                    if (invChild.GetType() == typeof(MemberReferenceExpression))
                    {
                        
                        if(invChild.Role.Equals(Roles.Argument))
                        {
                            ParseMemberReferenceExpression(invChild);
                            continue;
                        }
                        
                        MemberReferenceExpression mmbref = (MemberReferenceExpression)invChild;
                        foreach (AstNode memRefChild in invChild.Children)
                        {
                            if (memRefChild.GetType() == typeof(IdentifierExpression))
                            {
                                IdentifierExpression ie = (IdentifierExpression)memRefChild;
                                if (_currentIdentifiers.ContainsKey(ie.Identifier))
                                {
                                    Class methodInv = new Class(_currentIdentifiers[ie.Identifier]);
                                    _currentClass.AddReference(methodInv, ReferenceType.METHODINVOCATION, false, false, data.GetCurrentSourceFile());
                                    flag = true;
                                }
                                else
                                {
                                    Class staticMethodinv = new Class(ie.Identifier);
                                    _currentClass.AddReference(staticMethodinv, ReferenceType.STATICMETHODINVOCATION, false, false, data.GetCurrentSourceFile());
                                    flag = true;
                                }
                            }
                        }
                    }
                    else if (CheckChild(invChild))
                        continue;
                    else
                        RecursiveClassFinder(invChild);
                }
                return flag;
            }
            return false;
        }
        private bool ParseExpressionStatement(AstNode n)
        {
            if (n.GetType() == typeof(ExpressionStatement))
            {
                ExpressionStatement i = (ExpressionStatement)n;
                foreach (AstNode invChild in n.Children)
                {
                    if (invChild.GetType() == typeof(AssignmentExpression))
                    {
                        AssignmentExpression assref = (AssignmentExpression)invChild;
                        if (assref.Right.GetType() == typeof(ObjectCreateExpression))
                        {
                            foreach (AstNode memRefChild in invChild.Children)
                            {
                                if (memRefChild.GetType() == typeof(IdentifierExpression))
                                {
                                    IdentifierExpression ie = (IdentifierExpression)memRefChild;
                                    if (_currentIdentifiers.ContainsKey(ie.Identifier))
                                    {
                                        Class node = new Class(_currentIdentifiers[ie.Identifier]);
                                        _currentClass.AddReference(node, ReferenceType.CONSTRUCTORINVOCATION, false, false, data.GetCurrentSourceFile());
                                        return true;
                                    }
                                }
                            }
                        }

                    }
                    else if (CheckChild(invChild))
                        continue;
                    else
                        RecursiveClassFinder(invChild);
                }
            }
            return false;
        }
        private bool ParseMemberReferenceExpression(AstNode n)
        {
            if(n.GetType() == typeof(MemberReferenceExpression))
            {
                MemberReferenceExpression r = (MemberReferenceExpression)n;
                foreach(AstNode child in r.Children)
                {
                    if (child.GetType() == typeof(IdentifierExpression))
                    {
                        IdentifierExpression ie = (IdentifierExpression)child;
                        if (_currentIdentifiers.ContainsKey(ie.Identifier))
                        {
                            Class methodInv = new Class(_currentIdentifiers[ie.Identifier]);
                            CodeReference cr = new CodeReference(_currentClass, methodInv, ReferenceType.MEMBERREFERENCE, false, false, data.GetCurrentSourceFile());
                            foreach (AstNode idChild in r.Children)
                            {
                                if (idChild.GetType() == typeof(Identifier))
                                {
                                    Identifier id = (Identifier)idChild;
                                    cr.MemberProperty = id.Name;
                                }
                            }
                            _currentClass.AddReference(cr);
                        }
                        else
                        {
                            Class staticMethodinv = new Class(ie.Identifier);
                            _currentClass.AddReference(staticMethodinv, ReferenceType.STATICMEMBERREFERENCE, false, false, data.GetCurrentSourceFile());
                        }
                    }
                    else if (CheckChild(child))
                        continue;
                    else
                        RecursiveClassFinder(child);
                }
                return true;
            }
            return false;
        }
        private void RecursiveClassFinder(AstNode n)
        {
            foreach (AstNode child in n.Children)
            {
                if (CheckChild(child))
                    continue;
                else
                    RecursiveClassFinder(child);
            }
        }
        private bool CheckChild(AstNode child)
        {
            if (ParseVariableDeclarationStatement(child))
                return true;
            else if (ParseExpressionStatement(child))
                return true;
            else if (ParseMemberReferenceExpression(child))
                return true;
            else if (ParseInvocationExpression(child))
                return true;
            else if (ParseObjectCreateExpression(child))
                return true;
            else
            {
                return false;
            }
        }
        #endregion
        private string CheckProperty(AstNode n,string name)
        {
            foreach (PropertyInfo p in n.GetType().GetProperties(BindingFlags.Public | BindingFlags.Instance))
            {
                if (p.Name == "NodeType" || p.Name == "IsNull" || p.Name == "IsFrozen" || p.Name == "HasChildren")
                    continue;
                if (p.PropertyType == typeof(string) || p.PropertyType.IsEnum || p.PropertyType == typeof(bool))
                {
                    try
                    {
                        object val = p.GetValue(n, null);
                        if (p.Name.Equals(name) && val != null)
                        {
                            string ns = val.ToString();
                            return ns;
                        }
                    }
                    catch (TargetInvocationException)
                    {
                    }
                }
            }
            return null;
        }
        private class ParserStrings
        {
            public const string NAMESPACE = "NamespaceDeclaration";
            public const string USING = "UsingDeclaration";
            public const string TYPE = "TypeDeclaration";
            public const string MEMBERFIELD = "FieldDeclaration";
            public const string BLOCKSTMT = "BlockStatement";
            public const string METHOD = "MethodDeclaration";
            public const string METHODPARAMETER = "ParameterDeclaration";
            public const string SIMPLETYPE = "SimpleType";
            public const string MEMBERTYPE = "MemberType";
        }
        private class RoleStrings
        {
            public const string MEMBER = "Member";
            public const string TYPEMEMBER = "TypeMember";
            public const string METHODBODY = "Body";
            public const string PARAMETER = "Parameter";
            public const string TARGET = "Target";
            public const string BASETYPE = "BaseType";
            public const string TYPE = "Type";
        }
    }
}